/*
 * ztrace.h - definitions related to ztrace user-level-scheduler functionality
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef __ZTRACE_H
#define __ZTRACE_H

#include <revirt/rlog.h>

#define TRAP_FLAG (1<<8)

#define ZTRACE_SSTEP		1
#define ZTRACE_SSTEP_WAIT	2
#define ZTRACE_GETREG		3
#define ZTRACE_PUTREG		4
#define ZTRACE_ATTACH		5
#define ZTRACE_DETACH		6
#define ZTRACE_SYSCALL		7
#define ZTRACE_GETSTATE		8
#define ZTRACE_SYSCALL_ENTER	9 
#define ZTRACE_SYSCALL_EXIT	10
#define ZTRACE_SIGNALQUEUE	11
#define ZTRACE_RUNUNTIL		12
#define ZTRACE_COPYUSER		13
#define ZTRACE_XSTOP		14
#define ZTRACE_EXIT		15
#define ZTRACE_BLOCK		16
#define ZTRACE_DEBUGWAKE	17
#define ZTRACE_PEEK		18
#define ZTRACE_RESUME		19
#define ZTRACE_CLEARBRANCH	20
#define ZTRACE_PERFOVF		21
#define ZTRACE_SET_OPT		22

#define ZTRACE_NOSIG  		29
#define ZTRACE_HANG		30
#define ZTRACE_XSTOPTEST	31
#define ZTRACE_MARKER		32

#define ZTRACE_WATCH		40
#define ZTRACE_READWATCHBUFFER	41
#define ZTRACE_DELETEWATCH	42


/* volume 3. pages 484-486 */
#define ZTRC_DR6_B0	1

#define ZTRC_DR7_L0	1
#define ZTRC_DR7_LE	(1<<8)
#define ZTRC_DR7_RW0_EX 0
#define ZTRC_DR7_LEN0_0	0


struct ztrace_rununtil_struct {
	unsigned eip;
	unsigned long wchan;
	int flags;
	s64 count;
};

struct ztrace_state_struct {
#if 0
	unsigned long regEAX,
		      regEBX,
		      regECX,
		      regEDX,
		      regEFL,
		      regEIP,
		      regXSS;
#endif
	int replay_flags;
	unsigned char code[16];
	struct pt_regs regs;
	unsigned long debugreg[8];
	unsigned long cr2, trap_no, error_code;
	struct replay_ctrdata ctr;
	unsigned int wchan;
	int zt_flags;
	struct ztrace_rununtil_struct zt_rununtil;
	unsigned char pcode[16];
	s64 zt_perferr;
};

struct ztrace_breturn {
	unsigned eip;
	struct replay_ctrdata ctr;
	unsigned int wchan;
#if 0
	unsigned bcount;
	unsigned intrcount;
	unsigned excount;
#endif
};

struct ztrace_xuser_stop {
	int nrxuser;
	int edi;
};

struct ztrace_write_record {
	unsigned long len;
	long long offset;
};

#define ZTRACE_XUSER_STOP (1 << 0)
#define ZTRACE_SYSENTER_STOP (1 << 1)
#define ZTRACE_SYSEXIT_STOP (1 << 2)
#define ZTRACE_BKPT (1 << 3)
#define ZTRACE_BKPT_IN_KERNEL (1 << 4)
#define ZTRACE_WAITEXIT (1 << 5)
#define ZTRACE_BLOCK_STOP (1 << 6)
#define ZTRACE_SIGNAL_STEP (1 << 7)
#define ZTRACE_PERF (1 << 8)
#define ZTRACE_ACTMASK ( ~ (ZTRACE_XUSER_STOP|     \
			    ZTRACE_SYSENTER_STOP|  \
			    ZTRACE_SYSEXIT_STOP|   \
			    ZTRACE_BKPT |          \
			    ZTRACE_BLOCK_STOP |    \
			    ZTRACE_PERF |    \
			    ZTRACE_WAITEXIT ) )

#ifdef __KERNEL__


extern struct ztrace_xuser_stop zt_xstop;
extern int zt_flags;


int sys_ztrace(int command, int pid, unsigned long addr, unsigned long data, 
			unsigned int len);

#define zreg(tsk,reg)  (((struct pt_regs*) tsk->thread.esp0)-1)->reg
#define zoff(tsk,off) \
 (*(   ((unsigned long *) ((struct pt_regs*) tsk->thread.esp0) -1) +off))

int ztrace_read_long(struct task_struct *tsk,unsigned long addr,
                        unsigned long *ret);
int ztrace_get_reg(struct task_struct *tsk,unsigned long addr,
			unsigned long *val);
int ztrace_put_reg(struct task_struct *tsk,unsigned long addr,
                        unsigned long val);

void ztrace_debug_pause(void);

void ztrace_tsk_xuser_schedule(void);

#endif /* __KERNEL__ */

#endif /* __ZTRACE_H */
