/*
 * log.c - Functions associated with ReVirt logging system
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/linkage.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/tqueue.h>
#include <linux/slab.h>
#include <asm/perfctr.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/debugreg.h>
#include <asm/user.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>
#include <asm/unistd.h>

#include <asm/fixmap.h>
#include <asm/apic.h>
#include <asm/mtrr.h>
#include <asm/siginfo.h>

unsigned int replay_interrupts;
unsigned int replay_exceptions1;
unsigned int replay_exceptions2;

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
unsigned int replay_ecount[24];
#endif
int softints;
unsigned int replay_edi=0;
int last_uid;

/*
 * The performance counters are 48 and 64 bits for Intel and AMD, respectively;
 * we currently only use the first 32 bits.  4 billion branches between
 * schedules should be enough for anybody, but you never know.  We'll
 * grow when we need to.
 */
/*
 * FIXME -- inline, put in .h if necessary
 */
void copy_branchdata(struct task_struct *from, 
			    struct replay_ctrdata *to) {

	if(from == current) {
		//printk("copy_branchdata called on current!\n");
		save_perfctr_regs(from);
	}

	to->hw_branch= from->perfctr[0].ctr;
#ifdef REPLAY_DEBUG
	if(from->perfctr[0].ctr >= (1LL<<32)) {
		printk("copy_branchdata: warning: large count: 0x%llx\n",
			from->perfctr[0].ctr);
	}
#endif

#if  defined(CONFIG_MPENTIUM4)
        to->hw_intr=from->replay_int;
	to->ex_pf = from->replay_ex;
	to->ex_dev= from->replay_nrex;
#elif defined(CONFIG_MK6) || defined(CONFIG_MK7) \
	|| defined(CONFIG_M686) || defined(CONFIG_MPENTIUMIII)
	to->ex_dev=from->perfctr[1].ctr;
	to->ex_pf=from->replay_ex;
	to->hw_intr=from->replay_int;
#else
# error "Don't know what to copy for this architecture!"
#endif

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	do {
	    int i;
	    for(i=0; i<23; i++)
	    	to->exceptions[i]=replay_ecount[i];
	} while(0);
	//to->exceptions[23] = replay_exceptions2;
#endif
}

void clear_branchdata(struct task_struct *tsk) {
	tsk->perfctr[0].ctr=0;
	tsk->perfctr[1].ctr=0;
	tsk->replay_nrex=0;
	tsk->replay_ex=0;
	tsk->replay_int=0;
	softints=0;
	if(tsk == current) {
//		printk("clear_branchdata called on current!\n");
		restore_perfctr_regs(tsk);
	}
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	do { 
		int i; 
		for(i=0; i<20; i++)  {
    		    replay_ecount[i]=0; 
		}
	} while(0);
#endif
}
#if 0
static inline int get_stack_long(struct task_struct *task, int offset)
{
	unsigned char *stack;
	stack = (unsigned char *)task->thread.esp0;
	stack += offset;
	return (*((int *)stack));
}

static inline int put_stack_long(struct task_struct *task, int offset,
	unsigned long data)
{
	unsigned char * stack;

	stack = (unsigned char *) task->thread.esp0;
	stack += offset;
	*(unsigned long *) stack = data;
	return 0;
}
#endif


/*
 * Copied from arch/i386/kernel/process.c:get_wchan()
 */
extern void scheduling_functions_start_here(void);
extern void scheduling_functions_end_here(void);
#define first_sched     ((unsigned long) scheduling_functions_start_here)
#define last_sched      ((unsigned long) scheduling_functions_end_here)

unsigned int replay_getwchan(struct task_struct * prev) {
          unsigned long ebp, esp, eip;
          unsigned long stack_page;
          int count=0;

	  stack_page = (unsigned long) prev;
          esp=prev->thread.esp;
          if(!stack_page || esp < stack_page || esp > 8188+stack_page)
                count=16;
          ebp = * (unsigned long *) esp;
          do {
                if(ebp < stack_page || ebp > 8184+stack_page)
                        return 0;
                eip = *(unsigned long *) (ebp + 4);
                if(eip < first_sched || eip >= last_sched) {
                        return eip;
                }
                ebp = * (unsigned long *) ebp;
                        
          } while(count++<16);
	  return 0;
}

void save_perfctr(struct task_struct* tsk) {
        struct rlog_schedule_entry le;


	if(!(tsk->replay_flags & REPLAY_LOG)) {
	    printk("save_perfctr called for unlogged process. \n pid: 0x%x\n" \
	    	   " replay_flags: 0x%x\n", 
	    		tsk->pid, tsk->replay_flags);
	    BUG();
	} else {
	    //printk("save_perfctr: pid 0x%x\n", tsk->pid);
	}

	if(tsk->replay_flags & REPLAY_ZOMBIE)
		return;

	le.type=ENTRY_SCHED;

	copy_branchdata(tsk,&le.ctr);
        le.pid= tsk->pid;

	le.eip=zreg(tsk,eip);
	le.data=zreg(tsk,ecx);

	le.nrxuser = tsk->nrxuser;
	le.flags = tsk->replay_flags;
	le.flags |= REPLAY_TSK_SETSTATE(tsk->state);

	if(tsk->replay_flags & REPLAY_IN_COPY_USER) {
	  le.edi = replay_edi;
	  printk("save_perfctr() called in __copy_user_atom. edi==0x%x\n", 
	  		replay_edi);
	 }
	else
	  le.edi = 0;

	/*
	 * If we're blocked, find out where.  Code copied from
	 * arch/i386/kernel/process.c:get_wchan().  Note that the
	 * switch_to() macro has already stored the proper esp in the
	 * task struct, so the fact that tsk == current is OK.
	 *
	 * FIXME are we sure that tsk->state != TASK_RUNNING implies
	 * that save_perfctr() was called from __switch_to?
	 *
	 * OK, for now we look anytime process is in umode.  Will this break
	 * anything?
	 */
	le.wchan=0;

	if(tsk->state != TASK_RUNNING) {
#if 0
//	if(!(tsk->replay_flags & REPLAY_PROCESS_IN_UMODE)) {
	  unsigned long ebp, esp, eip;
	  unsigned long stack_page;
	  int count=0;
	  stack_page = (unsigned long) tsk;
	  esp=tsk->thread.esp;
	  if(!stack_page || esp < stack_page || esp > 8188+stack_page)
	  	count=16;
	  ebp = * (unsigned long *) esp;
	  do {
	  	if(ebp < stack_page || ebp > 8184+stack_page)
			break;
		eip = *(unsigned long *) (ebp + 4);
		if(eip < first_sched || eip >= last_sched) {
			le.wchan = eip;
			break;
		}
		ebp = * (unsigned long *) ebp;
			
	  } while(count++<16);
#endif
		le.wchan = replay_getwchan(tsk);
	} 

	clear_branchdata(tsk);

	tsk->nrsyscalls= 0;

	demon_write_log((char *)&le, le.type, sizeof(le));
}

#undef last_sched
#undef first_sched

#if 0
void record_fork(struct task_struct *parent,struct task_struct * child) {
        struct rlog_entry_fork le;

        le.type= ENTRY_FORK;
        le.pid= parent->pid;
        le.child=child->pid;

	/*
	 * All logs go to the same file now.
	 */

        demon_write_log((char *)&le, le.type, sizeof(le));
}
#endif

void log_exit(void) {
        struct rlog_exit_entry ee;

        ee.type= ENTRY_EXIT;
	ee.pid=current->pid;
	current->replay_flags |= REPLAY_ZOMBIE;
	demon_write_log((char *)&ee, ee.type, sizeof(ee));
}

asmlinkage void log_sysret(struct pt_regs  * regs) {
	struct rlog_entry_sysret se;

	se.type = ENTRY_SYSRET;
	se.pid = current->pid;
	se.syscall = current->current_syscall;  // sigreturn clobbers orig_eax
#ifdef REPLAY_WHOLE_PTREGS
	se.regs = *regs;
#else
	se.regs.eax = regs->eax;
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	copy_branchdata(current, &se.ctr);
#endif
	se.nrxuser=current->nrxuser;

	current->nrxuser=0;

	demon_write_log((char *)&se, se.type, sizeof(se));
}

int log_lastsig=0;

void log_sigframe(struct __kernel_sigframe * frame) {
	struct rlog_entry_sigframe e;

	e.type = ENTRY_SIGFRAME;
	e.pid = current->pid;
	e.frame = *frame;

	demon_write_log((char *)&e, e.type, sizeof(e));
}

void log_signal(siginfo_t * sig) {
#ifdef REPLAY_WHOLE_SIGINFO
	char d[sizeof(struct rlog_entry_signal) + sizeof(siginfo_t)];
#else
	struct rlog_entry_signal le_r;
#endif
	int i, size;
	struct rlog_entry_signal * le;
	siginfo_t * si;

#ifdef REPLAY_WHOLE_SIGINFO
	le = (struct rlog_entry_signal *)d;
	si = (siginfo_t *) (d + sizeof(struct rlog_entry_signal));

	size=sizeof(struct rlog_entry_signal) + sizeof(siginfo_t);
#else
	le = &le_r;
	size = sizeof(le_r);
#endif
	le->type = ENTRY_SIGNAL;
	le->pid = current->pid;

	/* XXX Should we lock current->sigmask_lock here? */

	le->pending_signal = current->pending.signal;
	le->blocked = current->blocked;

	/*
	 * We need to do this because force_siginfo may change SIG_IGN
	 * to SIG_DFL.
	 * 
	 * Note that there is no 0 signal; so all signals are shifted
	 * down one -- so sigaddset will change the i+1 back to i for
	 * action[0].
	 */

	if(current->sig) {
	    for(i=0; i<_NSIG_WORDS; i++) 
	    	le->dfl.sig[i] = 0;
	    for(i=0; i<_NSIG; i++) {
	    	if(current->sig->action[i].sa.sa_handler == SIG_DFL) {
#if 0
			printk("rlog: pid %d sig %d SIG_DFL\n", current->pid,
					i+1);
#endif
			sigaddset(&le->dfl, i+1);
		}
#if 0
		printk("rlog: le->dfl 0x%08x %08x\n", le->dfl.sig[0],
				le->dfl.sig[1]);
#endif
	    }
	} else {
		printk("rlog: Strange, log_signals called for zombie...\n");
	}

#ifdef REPLAY_WHOLE_SIGINFO
	*si = *sig;
	demon_write_log(d, le->type, size);
#else
	le->si_signo = sig->si_signo;
	le->si_errno = sig->si_errno;
	le->si_code = sig->si_code;
	le->_pad[0] = sig->_sifields._pad[0];
	le->_pad[1] = sig->_sifields._pad[1];
	demon_write_log((char *)le, le->type, size);
#endif
}

void log_rt_sigframe(struct __kernel_rt_sigframe * frame) {
	struct rlog_entry_rt_sigframe e;

	e.type = ENTRY_RT_SIGFRAME;
	e.pid = current->pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	e.frame = *frame;
#else
	e.sig = frame->sig;
	e.trapno = frame->uc.uc_mcontext.trapno;
#endif
	demon_write_log((char *)&e, e.type, sizeof(e));
}
