/*
 * rexecve.c - Replay log using the ReVirt system
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/file.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/perfctr.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <revirt/rlog.h>

#include <linux/slab.h>

/*
 * sys_rexecve() sets the replay flag of the process and
 * executes a new program. 
 * Synopsis:
 * rexecve( char *filename, char **argp, char **envp, 
 *          char *syscall_log_filename);
 */
asmlinkage int sys_rexecve(struct pt_regs regs)
{
	char * filename;
	int error;
	struct rlog_header rh;

        /* perhaps we should do some checking, for the newer libc. */
	printk("sys_rexecve is called\n");

	/* the log name */
	filename= getname((char*) regs.esi); 
	printk("rexecve: Attempting to open file %s\n", filename);
	error=demon_read_open(filename);
	printk("rexecve: demon_read_open returned %d\n", error);
	if(error < 0)
		goto out;
	printk("rexecve: success\n");

	putname(filename);

	current->replay_flags= REPLAY_REPLAY | REPLAY_SSTEP;
	current->replay_nrex= 0;

	error= -EINVAL;

	/* 
	 * Could we construct argv and env here instead of making
	 * the user program do it?
	 */


	perfctr_enable_counters();
	perfctr_clear_counters();

	clear_branchdata(current);
	softints=0;

	/* rest is from the original execve code */
	filename = getname((char *) regs.ebx);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out;
	error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, 
							&regs);
	if (error == 0)
		current->ptrace &= ~PT_DTRACE;

	return error;
out:
	putname(filename);
	current->replay_flags= 0;
	return error;
}

