#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <ctype.h>

#include <revirt/rlog.h>
#include <revirt/flags.h>

/*
 * So asm/siginfo.h will compile...
 */
typedef long clock_t;

#include <asm/siginfo.h>

#define BUFSIZE (1024 * 1024) 
unsigned char buf[BUFSIZE];

void print_top(struct rlog_header * h /*, struct rlog_entry_top * le*/) {
	printf("%u:Top (0x%x %d)\n", h->id, h->type, h->len);
}

void print_bc(struct replay_ctrdata * le) {
	int i;
	printf(" x:0x%llx ix:0x%x p:0x%x d:0x%x\n",
		le->hw_branch, 
		le->hw_intr,
		le->ex_pf,
		le->ex_dev);
	if(le->hw_intr != le->ex_dev) {
		printf(" ---Interrupt counts do not match---\n");
	}
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
#warning listing exceptions
	for(i=0; i<24; i++) {
		if(le->exceptions[i]>0)
			printf("  [%d] 0x%x\n", i, le->exceptions[i]);
	}
#endif
}

void print_sched(struct rlog_header *h, struct rlog_schedule_entry * le) {
	int i;
	printf("%u:Sched (0x%x %d) ", h->id, h->type, h->len);
	printf("pid:%5d flags:0x%08x ",
		le->pid,
		le->flags);
	if(le->flags & REPLAY_LOG)
		printf("REPLAY_LOG");
	if(le->flags & REPLAY_PROCESS_IN_UMODE)
		printf(" | REPLAY_PROCESS_IN_UMODE");
	if(le->flags & REPLAY_START)
		printf(" | REPLAY_START");
	if(le->flags & REPLAY_AT_EXIT)
		printf(" | REPLAY_AT_EXIT");
	if(le->flags & REPLAY_XUSER)
		printf(" | REPLAY_XUSER");
	if(le->flags & REPLAY_DATASQUASH)
		printf(" | REPLAY_DATASQUASH");
	printf("\n tsk->state: ");
	switch(REPLAY_TSK_GETSTATE(le->flags)) {
	  case 0: printf("TASK_RUNNING\n"); break;
	  case 1: printf("TASK_INTERUPTIBLE\n"); break;
	  case 2: printf("TASK_UNINTERUPTIBLE\n"); break;
	  case 4: printf("TASK_ZOMBIE\n"); break;
	  case 8: printf("TASK_STOPPED\n"); break;
	  default: printf("unknown\n");
	}
	print_bc(&le->ctr);
	printf(" eip:0x%x ecx:0x%x edi:0x%x nrxuser:0x%x wchan:0x%x\n", 
		le->eip,
		le->data,
		le->edi,
		le->nrxuser,
		le->wchan);
	if(h->len > sizeof(struct rlog_schedule_entry)){
	  for(i=sizeof(struct rlog_schedule_entry);
		i < h->len;
		i+=4) {
		printf("0x%x\n", *(int *)(buf+i));
	  }
	}
}

void print_sysret(struct rlog_header *h, struct rlog_entry_sysret * le) {
	printf("%u:Sysret (0x%x %d) ", h->id, h->type, h->len);
	printf("pid:%d syscall:%d nrxuser:%d ret: %d\n",
					le->pid, 
					le->syscall, 
					le->nrxuser,
					le->regs.eax);
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	print_bc(&le->ctr);
#endif
}

void print_signal(struct rlog_header *h, struct rlog_entry_signal * le) {
	printf("%u:Signal (0x%x %d) ", h->id, h->type, h->len);
	printf("pid:%5d signo:%2d errno:%2d\n",
		le->pid,
		le->si_signo,
		le->si_errno);
#if 0
	printf(" b: 0x%x %x", 
		le->blocked.sig[0],
		le->blocked.sig[1]);
	printf(" p: 0x%x %x",
		le->pending_signal.sig[0],
		le->pending_signal.sig[1]);
	printf(" d: 0x%x %x",
		le->dfl.sig[0],
		le->dfl.sig[1]);
	si = (struct siginfo *)(((char *)le)+sizeof(le->signal));
	printf("si_signo: %d si_errno: %d si_code: %d\n", 
			si->si_signo,
			si->si_errno,
			si->si_code);
	for(i=0; i<7; i++) {
		printf("%8x %8x %8x %8x\n",
			si->_sifields._pad[i],
			si->_sifields._pad[i+8],
			si->_sifields._pad[i+15],
			si->_sifields._pad[i+22]);
	}
	printf("%8x\n", si->_sifields._pad[7]);
#endif
}

void print_rt_sigframe(struct rlog_header *h, struct rlog_entry_rt_sigframe * le) {
	printf("%u:rt_sigframe (0x%x %d) ", h->id, h->type, h->len);
#ifndef REPLAY_WHOLE_SIGFRAME
	printf("pid:%5d sig:%2d trapno:%2d\n",
		le->pid,
		le->sig,
		le->trapno);
#else
	printf("[don't know how to print whole sigframe yet]\n");
#endif
}

int main(int argc, char * argv[]) {
	int fd;
	int ar[2], i, j, l;
	struct rlog_header rh;
	struct rlog_schedule_entry * le;
	struct siginfo * si;

	if(argc!=2) {
		fprintf(stderr, "Usage: %s logfile\n", argv[0]);
	}
	printf("Opening file %s\n", argv[1]);
	if((fd=open(argv[1], O_RDONLY))<0) {
		perror("opening file");
		exit(1);
	}
	while((l=read(fd, &rh, sizeof(rh)))>0) {
		union {
			struct rlog_entry_signal signal;
			struct rlog_put_user_entry put_user;
			struct rlog_copy_user_entry copy_user;
			struct rlog_clear_user_entry clear_user;
			struct rlog_schedule_entry sched;
			struct rlog_entry_sigframe sigframe;
			struct rlog_entry_rt_sigframe rt_sigframe;
		} * le = (__typeof(le))buf;

		if((l=read(fd, buf, rh.len))<rh.len) {
			if(l<0)
				perror("reading file");
			else
				fprintf(stderr, "Bad file format: wanted %d, got %d\n", rh.len, l);
			exit(1);
		}

		switch(rh.type) {
		    case ENTRY_TOP:
			print_top(&rh);
			break;
		    case ENTRY_SCHED:
			print_sched(&rh, (struct rlog_schedule_entry *)buf);
			break;
		    case ENTRY_FORK:
		     	printf("%u Fork (0x%x %d)\n", rh.id, rh.type, rh.len);
			break;
		    case ENTRY_EXIT:
		     	printf("%u Exit (0x%x %d)\n", rh.id, rh.type, rh.len);
			break;
		    case ENTRY_SIGNAL:
			print_signal(&rh, (struct rlog_entry_signal *)buf);
			break;
		    case ENTRY_PUT_USER:
		     	printf("put_user() log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_COPY_USER:
		     	printf("copy_user() log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_CLEAR_USER:
		     	printf("clear_user() log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_SYSRET:
			print_sysret(&rh, (struct rlog_entry_sysret *)buf);
			break;
		    case ENTRY_SIGFRAME:
		     	printf("Sigframe log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_RT_SIGFRAME:
			print_rt_sigframe(&rh, (struct rlog_entry_rt_sigframe *)buf);
			break;
		    default:
			printf("Other log: 0x%x %d\n", rh.type, rh.len);
		}
#if 0
		if(1) {

		       switch(rh.type) {
			  case ENTRY_SCHED:
				printf(" type:%d\tpid:%d\tflags:0x%x\n",
					le->sched.type, 
					le->sched.pid,
					le->sched.flags);
				if(le->sched.flags & REPLAY_LOG)
					printf(" REPLAY_LOG");
				if(le->sched.flags & REPLAY_PROCESS_IN_UMODE)
					printf(" | REPLAY_PROCESS_IN_UMODE");
				if(le->sched.flags & REPLAY_START)
					printf(" | REPLAY_START");
				if(le->sched.flags & REPLAY_AT_EXIT)
					printf(" | REPLAY_AT_EXIT");
				if(le->sched.flags & REPLAY_XUSER)
					printf(" | REPLAY_XUSER");
				if(le->sched.flags & REPLAY_DATASQUASH)
					printf(" | REPLAY_DATASQUASH");
				printf("\n process->state: ");
				switch(REPLAY_TSK_GETSTATE(le->sched.flags)) {
				  case 0: printf("TASK_RUNNING\n"); break;
				  case 1: printf("TASK_INTERUPTIBLE\n"); break;
				  case 2: printf("TASK_UNINTERUPTIBLE\n"); break;
				  case 4: printf("TASK_ZOMBIE\n"); break;
				  case 8: printf("TASK_STOPPED\n"); break;
				  default: printf("unknown\n");
				}
				printf(" x:0x%x ix:0x%x p:0x%x d:0x%x\n",
					le->sched.ctr.hw_branch, 
					le->sched.ctr.hw_intr,
					le->sched.ctr.ex_pf,
					le->sched.ctr.ex_dev);
				printf(" eip:0x%x ecx:0x%x edi:0x%x nrxuser:0x%x wchan:0x%x\n", 
					le->sched.eip,
					le->sched.data,
					le->sched.edi,
					le->sched.nrxuser,
					le->sched.wchan);
				if(rh.len > sizeof(struct rlog_schedule_entry)){
				  for(i=sizeof(struct rlog_schedule_entry);
					i < rh.len;
					i+=4) {
					printf("0x%x\n", *(int *)(buf+i));
				  }
				}
				break;
			  case ENTRY_SIGNAL:
				printf(" type:%d\tpid:%d\tsigno:%d\terrno:%d\n",
					le->signal.type,
					le->signal.pid,
					le->signal.si_signo,
					le->signal.si_errno);
				printf(" b: 0x%x %x\n", 
					le->signal.blocked.sig[0],
					le->signal.blocked.sig[1]);
				printf(" p: 0x%x %x\n",
					le->signal.pending_signal.sig[0],
					le->signal.pending_signal.sig[1]);
				printf(" d: 0x%x %x\n",
					le->signal.dfl.sig[0],
					le->signal.dfl.sig[1]);
				si = (struct siginfo *)(((char *)le)+sizeof(le->signal));
				printf("si_signo: %d si_errno: %d si_code: %d\n", 
						si->si_signo,
						si->si_errno,
						si->si_code);
				for(i=0; i<7; i++) {
					printf("%8x %8x %8x %8x\n",
						si->_sifields._pad[i],
						si->_sifields._pad[i+8],
						si->_sifields._pad[i+15],
						si->_sifields._pad[i+22]);
				}
				printf("%8x\n", si->_sifields._pad[7]);
				break;
			  case ENTRY_PUT_USER:
				printf(" type:%d\tpid:%d\n",
					le->put_user.type, 
					le->put_user.pid);
				printf(" x:0x%x ptr:0x%x s:0x%x r:0x%x\n",
					le->put_user.xx, 
					le->put_user.pptr, 
					le->put_user.ssize, 
					le->put_user.rretval);
				printf(" syscall:%d\n", 
					le->put_user.syscall);
				break;
			  case ENTRY_COPY_USER:
				printf(" type:%d\tpid:%d to:0x%x size:0x%x\n",
					le->copy_user.type, 
					le->copy_user.pid, 
					le->copy_user.tto, 
					le->copy_user.ssize);
				printf(" nrsyscalls:%d\tsyscall:%d\n", 
					le->copy_user.nrsyscalls, 
					le->copy_user.syscall);
				break;
			  case ENTRY_CLEAR_USER:
				printf(" type:%d\tpid:%d to:0x%x n:0x%x\n",
					le->clear_user.type, 
					le->clear_user.pid, 
					le->clear_user.to, 
					le->clear_user.n);
				printf(" syscall:%d\n", 
					le->clear_user.syscall);
				break;
			  case ENTRY_SYSRET:
				printf(" type:%d pid:%d syscall:%d nrxuser:%d ret: %d\n",
					le->sysret.type, 
					le->sysret.pid, 
					le->sysret.syscall, 
					le->sysret.nrxuser,
					le->sysret.regs.eax);
				break;
			  case ENTRY_SIGFRAME:
				printf(" type:%d\tpid:%d\tsig:%d\n", 
					le->sigframe.type,
					le->sigframe.pid,
					le->sigframe.frame.sig);
				break;
			  case ENTRY_RT_SIGFRAME:
				printf(" type:%d\tpid:%d\t", 
					le->rt_sigframe.type,
					le->rt_sigframe.pid);
#ifdef REPLAY_LOG_WHOLE_SIGFRAME
				printf(" siginfo.signo:%d .errno:%d .code:%d\n",
				    le->rt_sigframe.frame.info.si_signo,
				    le->rt_sigframe.frame.info.si_errno,
				    le->rt_sigframe.frame.info.si_code);
#else
				printf("\n");
#endif
				break;
		      }
		} 
#endif
	}
}
		
int printdata(char *buf, int l) {	
	int i, j;
		for(i=0; i<l; i+=16) {
		    printf(" %08x:", i);
		    for(j=i; j<i+16 && j<l; j++) {
			if (!(j%4) && j>0)
				printf(" ");
			printf("%02x ", (unsigned int)buf[j]);
                    }
		    if(j<i+16) {
			for(; j<i+16; j++) {
				if(!j%4)
					printf("  ");
				printf("   ");
			}
		    }
		    printf("  ");
		    for(j=i; j<i+16; j++) {
			if(isprint(buf[j]))
				fputc(buf[j], stdout);
			else
				fputc('.', stdout);
		    }
		    printf("\n");
		}
}
