#include <linux/assert.h>
#include <linux/sched.h>	/* needed by strncmp */


#define byte unsigned char
#define DEBUG_UMIDE_IRQ 0

#if DEBUG_UMIDE_IRQ
#define UMIDE_GEN_IRQ \
do { \
	printk("%s: IRQ %d\n", __FUNCTION__, umide_irq[nr]); \
	umide_gen_irq(nr); \
} while (0)			/* no ; */
#else				/* DEBUG_UMIDE_IRQ */
#define UMIDE_GEN_IRQ umide_gen_irq(nr)	/* no ; */
#endif				/* DEBUG_UMIDE_IRQ */

/* from <linux/cdrom.h> kernel 2.4.18 --tg 14:17 2002-05-22 */
#define GPMODE_CAPABILITIES_PAGE 0x2a

/* seek stuff */
# define SEEK_SET               0	/* Seek from beginning of file.  */
# define SEEK_CUR               1	/* Seek from current position.  */
# define SEEK_END               2	/* Seek from end of file.  */

/* from <linux/hdreg.h> kernel 2.4.18 --tg 14:21 2002-05-22 */
#define WIN_RESTORE             0x10
#define WIN_READ                0x20
#define WIN_WRITE               0x30
#define WIN_SPECIFY             0x91	/* set drive geometry translation */
#define WIN_IDENTIFY            0xEC	/* ask drive to identify itself */
#define WIN_SMART               0xb0	/* self monitoring and reporting */

/* power stuff */
#define WIN_STANDBYNOW1         0xE0
#define WIN_STANDBYNOW2         0x94
#define WIN_SLEEPNOW1           0xE6
#define WIN_SLEEPNOW2           0x99
#define WIN_CHECKPOWERMODE1     0xE5
#define WIN_CHECKPOWERMODE2     0x98
#define WIN_IDLEIMMEDIATE       0xE1	/* force drive to become "ready" */

/* stats */
#define ERR_STAT                0x01
#define INDEX_STAT              0x02
#define ECC_STAT                0x04	/* Corrected error */
#define DRQ_STAT                0x08
#define SEEK_STAT               0x10
#define WRERR_STAT              0x20
#define READY_STAT              0x40
#define BUSY_STAT               0x80

/* ATAPI stuff */
#define WIN_PIDENTIFY           0xA1	/* identify ATAPI device        */
#define WIN_SRST                0x08	/* ATAPI soft reset command */
#define WIN_PACKETCMD           0xA0	/* Send a packet command. */

/* some own constans */
#define UMIDE_BS                11 /* 2048 */
#define UMIDE_MAX               63488
#define UMIDE_BUSES             6

/* structure returned by HDIO_GET_IDENTITY, as per ANSI ATA2 rev.2f spec */
struct hd_driveid {
	unsigned short config;	/* lots of obsolete bit flags */
	unsigned short cyls;	/* "physical" cyls */
	unsigned short reserved2;	/* reserved (word 2) */
	unsigned short heads;	/* "physical" heads */
	unsigned short track_bytes;	/* unformatted bytes per track */
	unsigned short sector_bytes;	/* unformatted bytes per sector */
	unsigned short sectors;	/* "physical" sectors per track */
	unsigned short vendor0;	/* vendor unique */
	unsigned short vendor1;	/* vendor unique */
	unsigned short vendor2;	/* vendor unique */
	unsigned char serial_no[20];	/* 0 = not_specified */
	unsigned short buf_type;
	unsigned short buf_size;	/* 512 byte increments; 0 = not_specified */
	unsigned short ecc_bytes;	/* for r/w long cmds; 0 = not_specified */
	unsigned char fw_rev[8];	/* 0 = not_specified */
	unsigned char model[40];	/* 0 = not_specified */
	unsigned char max_multsect;	/* 0=not_implemented */
	unsigned char vendor3;	/* vendor unique */
	unsigned short dword_io;	/* 0=not_implemented; 1=implemented */
	unsigned char vendor4;	/* vendor unique */
	unsigned char capability;	/* bits 0:DMA 1:LBA 2:IORDYsw 3:IORDYsup */
	unsigned short reserved50;	/* reserved (word 50) */
	unsigned char vendor5;	/* vendor unique */
	unsigned char tPIO;	/* 0=slow, 1=medium, 2=fast */
	unsigned char vendor6;	/* vendor unique */
	unsigned char tDMA;	/* 0=slow, 1=medium, 2=fast */
	unsigned short field_valid;	/* bits 0:cur_ok 1:eide_ok */
	unsigned short cur_cyls;	/* logical cylinders */
	unsigned short cur_heads;	/* logical heads */
	unsigned short cur_sectors;	/* logical sectors per track */
	unsigned short cur_capacity0;	/* logical total sectors on drive */
	unsigned short cur_capacity1;	/*  (2 words, misaligned int)     */
	unsigned char multsect;	/* current multiple sector count */
	unsigned char multsect_valid;	/* when (bit0==1) multsect is ok */
	unsigned int lba_capacity;	/* total number of sectors */
	unsigned short dma_1word;	/* single-word dma info */
	unsigned short dma_mword;	/* multiple-word dma info */
	unsigned short eide_pio_modes;	/* bits 0:mode3 1:mode4 */
	unsigned short eide_dma_min;	/* min mword dma cycle time (ns) */
	unsigned short eide_dma_time;	/* recommended mword dma cycle time (ns) */
	unsigned short eide_pio;	/* min cycle time (ns), no IORDY  */
	unsigned short eide_pio_iordy;	/* min cycle time (ns), with IORDY */
	unsigned short words69_70[2];	/* reserved words 69-70 */
	/* HDIO_GET_IDENTITY currently returns only words 0 through 70 */
	unsigned short words71_74[4];	/* reserved words 71-74 */
	unsigned short queue_depth;	/*  */
	unsigned short words76_79[4];	/* reserved words 76-79 */
	unsigned short major_rev_num;	/*  */
	unsigned short minor_rev_num;	/*  */
	unsigned short command_set_1;	/* bits 0:Smart 1:Security 2:Removable 3:PM */
	unsigned short command_set_2;	/* bits 14:Smart Enabled 13:0 zero */
	unsigned short cfsse;	/* command set-feature supported extensions */
	unsigned short cfs_enable_1;	/* command set-feature enabled */
	unsigned short cfs_enable_2;	/* command set-feature enabled */
	unsigned short csf_default;	/* command set-feature default */
	unsigned short dma_ultra;	/*  */
	unsigned short word89;	/* reserved (word 89) */
	unsigned short word90;	/* reserved (word 90) */
	unsigned short CurAPMvalues;	/* current APM values */
	unsigned short word92;	/* reserved (word 92) */
	unsigned short hw_config;	/* hardware config */
	unsigned short words94_125[32];	/* reserved words 94-125 */
	unsigned short last_lun;	/* reserved (word 126) */
	unsigned short word127;	/* reserved (word 127) */
	unsigned short dlf;	/* device lock function
				 * 15:9        reserved
				 * 8   security level 1:max 0:high
				 * 7:6 reserved
				 * 5   enhanced erase
				 * 4   expire
				 * 3   frozen
				 * 2   locked
				 * 1   en/disabled
				 * 0   capability
				 */
	unsigned short csfo;	/* current set features options
				 * 15:4        reserved
				 * 3   auto reassign
				 * 2   reverting
				 * 1   read-look-ahead
				 * 0   write cache
				 */
	unsigned short words130_155[26];	/* reserved vendor words 130-155 */
	unsigned short word156;
	unsigned short words157_159[3];	/* reserved vendor words 157-159 */
	unsigned short words160_255[95];	/* reserved words 160-255 */
};

#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>

/* from <linux/ide.h> kernel 2.4.18 --tg 19:39 2002-05-20 */
#define ide_scsi        0x21
#define ide_disk        0x20
#define ide_optical     0x7
#define ide_cdrom       0x5
#define ide_tape        0x1
#define ide_floppy      0x0

/* little macro to don't write that all the time */
#define HDX (nr*2+unit+'a')

typedef enum {
	mechtype_caddy = 0,
	mechtype_tray = 1,
	mechtype_popup = 2,
	mechtype_individual_changer = 4,
	mechtype_cartridge_changer = 5
} mechtype_t;

struct mode_page_header {
	__u16 mode_data_length;
	__u8 medium_type;
	__u8 reserved1;
	__u8 reserved2;
	__u8 reserved3;
	__u16 desc_length;
};

struct atapi_capabilities_page {
	struct mode_page_header header;
#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 parameters_saveable:1;
	__u8 reserved1:1;
	__u8 page_code:6;
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	__u8 page_code:6;
	__u8 reserved1:1;
	__u8 parameters_saveable:1;
#else
#error "Please fix <asm/byteorder.h>"
#endif

	unsigned char page_length;

#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 reserved2:2;
	/* Drive supports reading of DVD-RAM discs */
	__u8 dvd_ram_read:1;
	/* Drive supports reading of DVD-R discs */
	__u8 dvd_r_read:1;
	/* Drive supports reading of DVD-ROM discs */
	__u8 dvd_rom:1;
	/* Drive supports reading CD-R discs with addressing method 2 */
	__u8 method2:1;		/* reserved in 1.2 */
	/* Drive can read from CD-R/W (CD-E) discs (orange book, part III) */
	__u8 cd_rw_read:1;	/* reserved in 1.2 */
	/* Drive supports read from CD-R discs (orange book, part II) */
	__u8 cd_r_read:1;	/* reserved in 1.2 */
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	/* Drive supports read from CD-R discs (orange book, part II) */
	__u8 cd_r_read:1;	/* reserved in 1.2 */
	/* Drive can read from CD-R/W (CD-E) discs (orange book, part III) */
	__u8 cd_rw_read:1;	/* reserved in 1.2 */
	/* Drive supports reading CD-R discs with addressing method 2 */
	__u8 method2:1;
	/* Drive supports reading of DVD-ROM discs */
	__u8 dvd_rom:1;
	/* Drive supports reading of DVD-R discs */
	__u8 dvd_r_read:1;
	/* Drive supports reading of DVD-RAM discs */
	__u8 dvd_ram_read:1;
	__u8 reserved2:2;
#else
#error "Please fix <asm/byteorder.h>"
#endif

#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 reserved3:2;
	/* Drive can write DVD-RAM discs */
	__u8 dvd_ram_write:1;
	/* Drive can write DVD-R discs */
	__u8 dvd_r_write:1;
	__u8 reserved3a:1;
	/* Drive can fake writes */
	__u8 test_write:1;
	/* Drive can write to CD-R/W (CD-E) discs (orange book, part III) */
	__u8 cd_rw_write:1;	/* reserved in 1.2 */
	/* Drive supports write to CD-R discs (orange book, part II) */
	__u8 cd_r_write:1;	/* reserved in 1.2 */
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	/* Drive can write to CD-R discs (orange book, part II) */
	__u8 cd_r_write:1;	/* reserved in 1.2 */
	/* Drive can write to CD-R/W (CD-E) discs (orange book, part III) */
	__u8 cd_rw_write:1;	/* reserved in 1.2 */
	/* Drive can fake writes */
	__u8 test_write:1;
	__u8 reserved3a:1;
	/* Drive can write DVD-R discs */
	__u8 dvd_r_write:1;
	/* Drive can write DVD-RAM discs */
	__u8 dvd_ram_write:1;
	__u8 reserved3:2;
#else
#error "Please fix <asm/byteorder.h>"
#endif

#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 reserved4:1;
	/* Drive can read multisession discs. */
	__u8 multisession:1;
	/* Drive can read mode 2, form 2 data. */
	__u8 mode2_form2:1;
	/* Drive can read mode 2, form 1 (XA) data. */
	__u8 mode2_form1:1;
	/* Drive supports digital output on port 2. */
	__u8 digport2:1;
	/* Drive supports digital output on port 1. */
	__u8 digport1:1;
	/* Drive can deliver a composite audio/video data stream. */
	__u8 composite:1;
	/* Drive supports audio play operations. */
	__u8 audio_play:1;
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	/* Drive supports audio play operations. */
	__u8 audio_play:1;
	/* Drive can deliver a composite audio/video data stream. */
	__u8 composite:1;
	/* Drive supports digital output on port 1. */
	__u8 digport1:1;
	/* Drive supports digital output on port 2. */
	__u8 digport2:1;
	/* Drive can read mode 2, form 1 (XA) data. */
	__u8 mode2_form1:1;
	/* Drive can read mode 2, form 2 data. */
	__u8 mode2_form2:1;
	/* Drive can read multisession discs. */
	__u8 multisession:1;
	__u8 reserved4:1;
#else
#error "Please fix <asm/byteorder.h>"
#endif

#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 reserved5:1;
	/* Drive can return Media Catalog Number (UPC) info. */
	__u8 upc:1;
	/* Drive can return International Standard Recording Code info. */
	__u8 isrc:1;
	/* Drive supports C2 error pointers. */
	__u8 c2_pointers:1;
	/* R-W data will be returned deinterleaved and error corrected. */
	__u8 rw_corr:1;
	/* Subchannel reads can return combined R-W information. */
	__u8 rw_supported:1;
	/* Drive can continue a read cdda operation from a loss of streaming. */
	__u8 cdda_accurate:1;
	/* Drive can read Red Book audio data. */
	__u8 cdda:1;
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	/* Drive can read Red Book audio data. */
	__u8 cdda:1;
	/* Drive can continue a read cdda operation from a loss of streaming. */
	__u8 cdda_accurate:1;
	/* Subchannel reads can return combined R-W information. */
	__u8 rw_supported:1;
	/* R-W data will be returned deinterleaved and error corrected. */
	__u8 rw_corr:1;
	/* Drive supports C2 error pointers. */
	__u8 c2_pointers:1;
	/* Drive can return International Standard Recording Code info. */
	__u8 isrc:1;
	/* Drive can return Media Catalog Number (UPC) info. */
	__u8 upc:1;
	__u8 reserved5:1;
#else
#error "Please fix <asm/byteorder.h>"
#endif

#if defined(__BIG_ENDIAN_BITFIELD)
	/* Drive mechanism types. */
	mechtype_t mechtype:3;
	__u8 reserved6:1;
	/* Drive can eject a disc or changer cartridge. */
	__u8 eject:1;
	/* State of prevent/allow jumper. */
	__u8 prevent_jumper:1;
	/* Present state of door lock. */
	__u8 lock_state:1;
	/* Drive can lock the door. */
	__u8 lock:1;
#elif defined(__LITTLE_ENDIAN_BITFIELD)

	/* Drive can lock the door. */
	__u8 lock:1;
	/* Present state of door lock. */
	__u8 lock_state:1;
	/* State of prevent/allow jumper. */
	__u8 prevent_jumper:1;
	/* Drive can eject a disc or changer cartridge. */
	__u8 eject:1;
	__u8 reserved6:1;
	/* Drive mechanism types. */
	mechtype_t mechtype:3;
#else
#error "Please fix <asm/byteorder.h>"
#endif

#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 reserved7:4;
	/* Drive supports software slot selection. */
	__u8 sss:1;		/* reserved in 1.2 */
	/* Changer can report exact contents of slots. */
	__u8 disc_present:1;	/* reserved in 1.2 */
	/* Audio for each channel can be muted independently. */
	__u8 separate_mute:1;
	/* Audio level for each channel can be controlled independently. */
	__u8 separate_volume:1;
#elif defined(__LITTLE_ENDIAN_BITFIELD)

	/* Audio level for each channel can be controlled independently. */
	__u8 separate_volume:1;
	/* Audio for each channel can be muted independently. */
	__u8 separate_mute:1;
	/* Changer can report exact contents of slots. */
	__u8 disc_present:1;	/* reserved in 1.2 */
	/* Drive supports software slot selection. */
	__u8 sss:1;		/* reserved in 1.2 */
	__u8 reserved7:4;
#else
#error "Please fix <asm/byteorder.h>"
#endif

	/* Note: the following four fields are returned in big-endian form. */
	/* Maximum speed (in kB/s). */
	unsigned short maxspeed;
	/* Number of discrete volume levels. */
	unsigned short n_vol_levels;
	/* Size of cache in drive, in kB. */
	unsigned short buffer_size;
	/* Current speed (in kB/s). */
	unsigned short curspeed;
	char pad[4];
};

/* Structure of a MSF cdrom address. */
struct atapi_msf {
	unsigned char reserved;
	unsigned char minute;
	unsigned char second;
	unsigned char frame;
};

/* Space to hold the disk TOC. */
#define MAX_TRACKS 99
struct atapi_toc_header {
	unsigned short toc_length;
	unsigned char first_track;
	unsigned char last_track;
};

struct atapi_toc_entry {
	unsigned char reserved1;
#if defined(__BIG_ENDIAN_BITFIELD)
	__u8 adr:4;
	__u8 control:4;
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	__u8 control:4;
	__u8 adr:4;
#else
#error "Please fix <asm/byteorder.h>"
#endif
	unsigned char track;
	unsigned char reserved2;
	union {
		unsigned lba;
		struct atapi_msf msf;
	} addr;
};

struct atapi_toc {
	int last_session_lba;
	int xa_flag;
	unsigned long capacity;
	struct atapi_toc_header hdr;
	struct atapi_toc_entry ent[MAX_TRACKS + 1];
	/* One extra for the leadout. */
};

#define CD_SECS              60	/* seconds per minute */
#define CD_FRAMES            75	/* frames per second */
#define CD_MSF_OFFSET       150	/* MSF numbering offset of first frame */

void
lba_to_msf (int lba, byte * m, byte * s, byte * f)
{
	lba += CD_MSF_OFFSET;
	lba &= 0xffffff;	/* negative lbas use only 24 bits */
	*m = lba / (CD_SECS * CD_FRAMES);
	lba %= (CD_SECS * CD_FRAMES);
	*s = lba / CD_FRAMES;
	*f = lba % CD_FRAMES;
}

int
msf_to_lba (byte m, byte s, byte f)
{
	return (((m * CD_SECS) + s) * CD_FRAMES + f) - CD_MSF_OFFSET;
}

#define ARY_LEN(a) ((sizeof(a) / sizeof(a[0])))

/*********************************************************************
  Generic Packet commands, MMC commands, and such
 *********************************************************************/

/* The generic packet command opcodes for CD/DVD Logical Units,
 * From Table 57 of the SFF8090 Ver. 3 (Mt. Fuji) draft standard. */
#define GPCMD_BLANK                         0xa1
#define GPCMD_CLOSE_TRACK                   0x5b
#define GPCMD_FLUSH_CACHE                   0x35
#define GPCMD_FORMAT_UNIT                   0x04
#define GPCMD_GET_CONFIGURATION             0x46
#define GPCMD_GET_EVENT_STATUS_NOTIFICATION 0x4a
#define GPCMD_GET_PERFORMANCE               0xac
#define GPCMD_INQUIRY                       0x12
#define GPCMD_LOAD_UNLOAD                   0xa6
#define GPCMD_MECHANISM_STATUS              0xbd
#define GPCMD_MODE_SELECT_10                0x55
#define GPCMD_MODE_SENSE_10                 0x5a
#define GPCMD_PAUSE_RESUME                  0x4b
#define GPCMD_PLAY_AUDIO_10                 0x45
#define GPCMD_PLAY_AUDIO_MSF                0x47
#define GPCMD_PLAY_AUDIO_TI                 0x48
#define GPCMD_PLAY_CD                       0xbc
#define GPCMD_PREVENT_ALLOW_MEDIUM_REMOVAL  0x1e
#define GPCMD_READ_10                       0x28
#define GPCMD_READ_12                       0xa8
#define GPCMD_READ_CDVD_CAPACITY            0x25
#define GPCMD_READ_CD                       0xbe
#define GPCMD_READ_CD_MSF                   0xb9
#define GPCMD_READ_DISC_INFO                0x51
#define GPCMD_READ_DVD_STRUCTURE            0xad
#define GPCMD_READ_FORMAT_CAPACITIES        0x23
#define GPCMD_READ_HEADER                   0x44
#define GPCMD_READ_TRACK_RZONE_INFO         0x52
#define GPCMD_READ_SUBCHANNEL               0x42
#define GPCMD_READ_TOC_PMA_ATIP             0x43
#define GPCMD_REPAIR_RZONE_TRACK            0x58
#define GPCMD_REPORT_KEY                    0xa4
#define GPCMD_REQUEST_SENSE                 0x03
#define GPCMD_RESERVE_RZONE_TRACK           0x53
#define GPCMD_SCAN                          0xba
#define GPCMD_SEEK                          0x2b
#define GPCMD_SEND_DVD_STRUCTURE            0xad
#define GPCMD_SEND_EVENT                    0xa2
#define GPCMD_SEND_KEY                      0xa3
#define GPCMD_SEND_OPC                      0x54
#define GPCMD_SET_READ_AHEAD                0xa7
#define GPCMD_SET_STREAMING                 0xb6
#define GPCMD_START_STOP_UNIT               0x1b
#define GPCMD_STOP_PLAY_SCAN                0x4e
#define GPCMD_TEST_UNIT_READY               0x00
#define GPCMD_VERIFY_10                     0x2f
#define GPCMD_WRITE_10                      0x2a
#define GPCMD_WRITE_AND_VERIFY_10           0x2e
/* This is listed as optional in ATAPI 2.6, but is (curiously) 
   missing from Mt. Fuji, Table 57.  It _is_ mentioned in Mt. Fuji
 * Table 377 as an MMC command for SCSi devices though...  Most ATAPI
 * drives support it. */
#define GPCMD_SET_SPEED                     0xbb
/* This seems to be a SCSI specific CD-ROM opcode 
 * to play data at track/index */
#define GPCMD_PLAYAUDIO_TI                  0x48
/*
 * From MS Media Status Notification Support Specification. For
 * older drives only.
 */
#define GPCMD_GET_MEDIA_STATUS              0xda

const struct {
	unsigned short packet_command;
	const char *const text;
} umide_gpcmd_texts[] = {
	{ GPCMD_TEST_UNIT_READY, "Test Unit Ready"},
	{ GPCMD_REQUEST_SENSE, "Request Sense"},
	{ GPCMD_FORMAT_UNIT, "Format Unit"},
	{ GPCMD_INQUIRY, "Inquiry"},
	{ GPCMD_START_STOP_UNIT, "Start/Stop Unit"},
	{ GPCMD_PREVENT_ALLOW_MEDIUM_REMOVAL, "Prevent/Allow Medium Removal"},
	{ GPCMD_READ_FORMAT_CAPACITIES, "Read Format Capacities"},
	{ GPCMD_READ_CDVD_CAPACITY, "Read Cd/Dvd Capacity"},
	{ GPCMD_READ_10, "Read 10"},
	{ GPCMD_WRITE_10, "Write 10"},
	{ GPCMD_SEEK, "Seek"},
	{ GPCMD_WRITE_AND_VERIFY_10, "Write and Verify 10"},
	{ GPCMD_VERIFY_10, "Verify 10"},
	{ GPCMD_FLUSH_CACHE, "Flush Cache"},
	{ GPCMD_READ_SUBCHANNEL, "Read Subchannel"},
	{ GPCMD_READ_TOC_PMA_ATIP, "Read Table of Contents"},
	{ GPCMD_READ_HEADER, "Read Header"},
	{ GPCMD_PLAY_AUDIO_10, "Play Audio 10"},
	{ GPCMD_GET_CONFIGURATION, "Get Configuration"},
	{ GPCMD_PLAY_AUDIO_MSF, "Play Audio MSF"},
	{ GPCMD_PLAYAUDIO_TI, "Play Audio TrackIndex"},
	{ GPCMD_GET_EVENT_STATUS_NOTIFICATION, "Get Event Status Notification"},
	{ GPCMD_PAUSE_RESUME, "Pause/Resume"},
	{ GPCMD_STOP_PLAY_SCAN, "Stop Play/Scan"},
	{ GPCMD_READ_DISC_INFO, "Read Disc Info"},
	{ GPCMD_READ_TRACK_RZONE_INFO, "Read Track Rzone Info"},
	{ GPCMD_RESERVE_RZONE_TRACK, "Reserve Rzone Track"},
	{ GPCMD_SEND_OPC, "Send OPC"},
	{ GPCMD_MODE_SELECT_10, "Mode Select 10"},
	{ GPCMD_REPAIR_RZONE_TRACK, "Repair Rzone Track"},
	{ GPCMD_MODE_SENSE_10, "Mode Sense 10"},
	{ GPCMD_CLOSE_TRACK, "Close Track"},
	{ GPCMD_BLANK, "Blank"},
	{ GPCMD_SEND_EVENT, "Send Event"},
	{ GPCMD_SEND_KEY, "Send Key"},
	{ GPCMD_REPORT_KEY, "Report Key"},
	{ GPCMD_LOAD_UNLOAD, "Load/Unload"},
	{ GPCMD_SET_READ_AHEAD, "Set Read-ahead"},
	{ GPCMD_READ_12, "Read 12"},
	{ GPCMD_GET_PERFORMANCE, "Get Performance"},
	{ GPCMD_SEND_DVD_STRUCTURE, "Send DVD Structure"},
	{ GPCMD_READ_DVD_STRUCTURE, "Read DVD Structure"},
	{ GPCMD_SET_STREAMING, "Set Streaming"},
	{ GPCMD_READ_CD_MSF, "Read CD MSF"},
	{ GPCMD_SCAN, "Scan"},
	{ GPCMD_SET_SPEED, "Set Speed"},
	{ GPCMD_PLAY_CD, "Play CD"},
	{ GPCMD_MECHANISM_STATUS, "Mechanism Status"},
	{ GPCMD_READ_CD, "Read CD"}
};

#define UNIT0 0
#define UNIT1 1

/* FIXME: wrapper functions to avoid egcs compiler bug */
static int
lotol_read_wrapper (int fd, char *buf, int count)
{
	return lotol_read (fd, buf, count);
}

static int
lotol_write_wrapper (int fd, char *buf, int count)
{
	return lotol_write (fd, buf, count);
}

static int
lotol_lseek_wrapper (int fd, int offset, int whence)
{
	return lotol_lseek (fd, offset, whence);
}

static int
lotol__llseek_wrapper (uint fd, ulong hi, ulong lo, loff_t * res, uint whence)
{
	return lotol__llseek (fd, hi, lo, res, whence);
}

/* FIXME: configurable */
static int umide_irq[UMIDE_BUSES] = { 14, 15, 11, 10, 8, 12 };

#define SECTOR_SIZE 512

struct {
	struct {
		int fd;
		unsigned long phys_cyls;
		unsigned long phys_heads;
		unsigned long phys_sectors;
	} unit[2];

	unsigned char buf[UMIDE_MAX+SECTOR_SIZE];	/* FIXME Why? --tg 19:33 2002-05-20 */
	unsigned short head;
	unsigned short tail;
	unsigned int count;	/* 32 bit -> Ueberlauf gefaehrdet */
	unsigned int m2t;	/* 32 bit -> Ueberlauf gefaehrdet */
	unsigned char error;	/* error reg */
	unsigned char nsector;	/* sector count reg */
	unsigned char sector;	/* sector reg */
	unsigned char lcyl;	/* low cylinder reg */
	unsigned char hcyl;	/* high cylinder reg */
	unsigned char select;	/* select reg */
	unsigned char status;	/* status reg */
	unsigned char command;	/* command reg */
	unsigned char control;	/* device control reg */
	unsigned short pcdone;	/* for packedcmd_step ... is this packetcmd done? */
	unsigned short identify;	/* generate irq? */
	unsigned short active;	/* 0x01 = master; 0x02 = slave present */
} ide[UMIDE_BUSES] = {		/* ide[0-5].unit[0].fd = -1 */
	{ { { -1 } } },
	{ { { -1 } } },
	{ { { -1 } } },
	{ { { -1 } } },
	{ { { -1 } } },
	{ { { -1 } } }
};

/*
   known side effect: I initalize every array to -1 to don't get it in bss
   data. This is because I want't to initalize 'active' in the early startup
   phase via umide_init. All bss data cleared after this initaliziation so that
   this would _not_ have a effect, if they were bss data. So I _have_ to
   initalize them to != 0 ... --tg 15:46 2002-06-05
   */
