/* $Id: fi.c,v 1.2 2002/12/16 22:04:05 kingst Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <linux/net.h>
#include <sys/ptrace.h>
#include <sys/time.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/user.h>
#include <netinet/in.h>

#include "fi_defs.h"
#include "fi_interface.h"


/***********************************************************************
 * file global variables
 **********************************************************************/
fi_fltd_t *fi_afdl_first = NULL;
fi_fltd_t *fi_afdl_last = NULL;
fi_fltd_t *fi_first = NULL;
fi_fltd_t *fi_last = NULL;
fi_flt_inj_t fi_ = {
	0, /* pid */
	0, /* watch_enter */
	0, /* watch_return */
	0, /* num_fltd */
	NULL, /* watch_open_ptr */
	0, /* netinit */
	0, /* fd */
	0, /* syscall */
	0,  /* socketcall */
	};
int fi_socket = -1; /* communication with frontend */
int fi_node = -1; /* node-id of this node as used by frontend */
/* extern fi_sbuf_t fi_sbuf;*/ /* fi_parse.c */

/***********************************************************************
 * function prototypes
 ***********************************************************************
 * file internal functions 
 **********************************************************************/
void init_netinfo() ;
/*
void fi_init_global_vars(int);
*/

int fi_init_socket() ;
void fi_setsocket_async(int) ;
void fi_setsocket_sync(int) ;
/* FIXME: not needed.
void fi_setsocket_block(int) ;
*/

/***********************************************************************
 * exported functions
 **********************************************************************/
void fi_init(int, int);
void fi_exit();
void fi_tick();

/***********************************************************************
 * imported functions
 **********************************************************************/
extern void fi_dprintf(char *, ...); /* fi_print.c */
extern void fi_eprintf(char *, ...); /* fi_print.c */
extern void fi_print_all_dbg(fi_fltd_t *, char *); /* fi_print.c */
extern int parse_fltd(int, fi_fltd_t **); /* fi_parse.c */
extern int read_decnum (int, unsigned long*); /* fi_parse.c */
extern void activate_fltd(fi_fltd_t *) ; /* fi_inject.c */

/***********************************************************************
 * init_netinfo
 ***********************************************************************
 * Input:     none
 * Output:    none
 * Effects:   initializes fi_.netinfo[] and fi_.netinit. fi_.netinit will be set
              to the number of available network interfaces. Each time, an
	      interface is initialized (watch_bind_enter()), fi_.netinit will be
	      decreased, so we know when we are done.
 * Called By: fi_init_global_vars
 **********************************************************************/
void
init_netinfo()
{
	char s[128];
	int ret ;
	unsigned int myself, i, j;
	unsigned short port ;
	FILE *f;
	/* Parse the device files for the network devices. These files contain
	 * lines with ip-addr:port information for the sockets implementing the
	 * network devices. We identify the network device by the port it uses.
	 * For format of these files refer to UMLinux/userguide/usage.
	 * Right now, only network uses bind for initialization, console 
	 * uses connect to connect to port bound by frontend.
	 */
	fi_.netinit = 0 ; 
	for(i=0; i<MAX_ETH_CARDS; i++) {
		(fi_.netinfo[i]).device = NULL ; /* we don't need name */
		(fi_.netinfo[i]).port = 0 ;
		/* FIXME: now using fixed file descriptors! */
		(fi_.netinfo[i]).fd = -1 ; /* will be set later */
		snprintf(s, 8, "eth%d", i) ;
		f = fopen(s, "r");
		if (f !=  NULL){
			fscanf(f," %x\n",  &myself);
			assert(0<=myself && myself < MAX_ETH_CARDS);
			for(j=0;j<MAX_ETH_CARDS;j++){
				ret=fscanf(f," %*x:%*x:%*x:%*x:%*x:%*x %*x:%hx\n", &port);
				if (j == myself) break;
			}
			(fi_.netinfo[i]).port = port ; /* found port */
			fclose(f);
			fi_.netinit ++ ; /* increase number of valid nics */
		}
	}
	/* do same for console, so that we can later safely ignore the
	 * console-socketdescriptor on reads/writes.
	 */
	f = fopen("console", "r") ;
	(fi_.console).device = NULL ;
	(fi_.console).port = 0 ;
	(fi_.console).fd = -1 ; /* will be set later */
	if (f !=  NULL){
		ret = fscanf(f,"%*x:%hx\n", &port);
		(fi_.console).port = port ; /* found port */
		fclose(f);
	}
}
/***********************************************************************
 * fi_init_global_vars
 ***********************************************************************
 * Input:     int pid  =  process id of UMLinux virtual machine process
 * Output:    none
 * Effects:   initializes all file global variables if this tracer is
 *            (re)started (not called on restart of virtual machine).
 * Called By: fi_init
 **********************************************************************/
 /*
void
fi_init_global_vars()
{
	memset(fi_.fdl, 0, sizeof(fi_fltd_t)*(FI_MAX_FLTD));
	fi_.num_fltd       = 0;
	fi_.watch_return   = 0;
	memset(fi_.active, 0, sizeof(int)*(FI_MAX_FLT_LOC));
	fi_.watch_open_ptr = 0;
	memset(fi_.fds, 0, sizeof(fi_fltd_t*)*(FI_MAX_FLTD));
	fi_.fd             = -1;
	fi_.syscall        = -1;
	fi_.socketcall     = -1;
}
*/
void
fi_init_netinfo()
{
	/* we must watch for bind system calls and make a list of fd<->port in case
	 * we later get requests from frontend to inject network faults (bind is
	 * only called at bootup of virtual machine and there is no way later to
	 * find out which socket represents which network card).
	 */
	fi_.watch_enter   |= FI_WATCH_BIND | FI_WATCH_CONNECT ;
	/* make and save a list of devicenames and port-numbers. socket descriptors
	 * will be added later by watching bind system call
	 */
	init_netinfo() ;
}
/***********************************************************************
 * fi_setsocket_sync
 ***********************************************************************
 * Input:     int socket  =  socket descriptor
 * Output:    none
 * Effects:   configures socket for synchronous blocking IO
 * Called By: fi_tick
 **********************************************************************/
void
fi_setsocket_sync(int socket)
{
	int flags;
	int ret;
	/* Prepare socket for synchronous IO
	 */
	flags = fcntl(socket, F_GETFL, 0) ;
	flags &= ~O_ASYNC;
	flags &= ~O_NONBLOCK;
	ret = fcntl(socket, F_SETFL, flags);
	if (ret < 0) { _exit(1); }
}
/***********************************************************************
 * fi_setsocket_async
 ***********************************************************************
 * Input:     int socket  =  socket descriptor
 * Output:    none
 * Effects:   configures socket for asynchronous non-blocking IO
 * Called By: fi_tick
 **********************************************************************/
void
fi_setsocket_async(int socket)
{
	int flags;
	int ret;
	/* Prepare socket for asynchronous and nonblocking IO
	 * Set receiver of SIGIO (=self).
	 */
	flags = fcntl(socket, F_GETFL, 0) ;
	flags |= O_ASYNC;
	flags |= O_NONBLOCK;
	ret = fcntl(socket, F_SETFL, flags);
	if (ret < 0) { _exit(2); }
	ret = fcntl(socket, F_SETOWN, getpid());
	if (ret < 0) { _exit(2); }
	/* ret = fcntl(socket, F_SETSIG, 0) ; *//* 0 = default (SIGIO) */
	/* tracer.c sets up interupt-handler which calls fi_tick() */
}
/***********************************************************************
 * fi_init_socket
 ***********************************************************************
 * Input:     none
 * Output:    socket-descriptor initialized with information from file "tracer"
 * Effects:   
 * Called By: fi_init
 **********************************************************************/
int
fi_init_socket()
{
	FILE *f;
	unsigned long ipaddr;
	unsigned short ipport;
	struct sockaddr_in addr;
	int sock ;
	int ret;

	/* parse file */
	f = fopen("tracer", "r") ;
	if (f !=  NULL){
		ret = fscanf(f," %lx:%hx\n",  &ipaddr, &ipport);
		if (ret != 2) {
			fi_eprintf("%s: %s: %s.\n", "fi", "tracer", "bad
				syntax");
			exit(1);
		}
		fclose(f) ;
	} else {
		fi_eprintf("%s: socket info: %s.\n", "fi", strerror(errno));
		exit(1);
	}

	/* create socket */
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		fi_eprintf("tracer: cannot create socket\n") ;
		exit(5);
	}
	ret = fcntl(sock, F_SETFD, FD_CLOEXEC);
	assert(0==ret) ;
	/*
	fi_eprintf(stderr, "XXX1\n");fflush(stderr);
	fi_setsocket_async(sock) ;
	fi_eprintf("XXX2\n");fflush(stderr);
	*/

	/* connect with experiment controller */
	/* now this will only work, if file "tracer" contains IP address of real
	 * host running the experiment controller!
	fi_eprintf("XXX3\n");fflush(stderr);
	fi_eprintf("XXX4\n");fflush(stderr);
	 */
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(ipaddr);
	addr.sin_port = htons(ipport);
	ret = connect(sock, (struct sockaddr *) &addr, sizeof(addr));
	if (ret < 0) {
		fi_eprintf("tracer: cannot connect  socket ip %d port %d: %s\n",
			   ipaddr,ipport,strerror(errno)) ;
		exit(5);
	}
	fi_eprintf("%s:%d: Using socket %x\n", __FUNCTION__, fi_node, ipport) ;
	/* socket stays in synchronous mode until we read END_OF_FAULTLOAD.
	 * Then we set it to async until we receive an interrupt, when we will
	 * set it to sync again, until we read END_OF_FAULTLOAD.
	 */
	 return(sock) ;
}
/***********************************************************************
 * functions for interfacing with the tracer.
 **********************************************************************/
/***********************************************************************
 * fi_exit
 ***********************************************************************
 * Input:     none
 * Output:    none
 * Effects:   shuts down fi_socket
 * Called By: tracer in tracer.c
 **********************************************************************/
void
fi_exit()
{
	int ret ;
	int socket = fi_socket ;

	fi_eprintf("%s:%d %s: shutting down socket.\n",
			__FUNCTION__, fi_node, "fi");
	/* FIXME DBG
	*/
	if (0 <= socket) {
		ret = shutdown(socket, 2);
		if (ret < 0) {
			fi_eprintf("%s: WARNING: shutdown socket: %s.\n",
					__FUNCTION__, strerror(errno));
		}
	}
	if (0 <= socket) {
		ret = close(socket) ;
		if (ret < 0) {
			fi_eprintf("%s: WARNING: close: %s.\n",
					__FUNCTION__, strerror(errno));
		}
	}
	fi_print_all_dbg(fi_afdl_first, "fi.alldbg.exit");
	/* FIXME DBG
	*/
}
/***********************************************************************
 * fi_init
 ***********************************************************************
 * Input:     int pid  =  process id of UMLinux virtual machine process
 * Output:    none
 * Effects:   intializes fault injector
 * Called By: tracer in tracer.c
 **********************************************************************/
void
fi_init(int pid, int ncpu)
{
	fi_dprintf(0);
	fi_eprintf(0);
	fi_eprintf("%s pid=%d\n", __FUNCTION__,pid);fflush(stderr);
	if (pid == 0) { /* new start of lilo/tracer */
		memset(fi_.fdl, 0, sizeof(fi_fltd_t)*(FI_MAX_FLTD));
		fi_.num_fltd       = 0;
		fi_.watch_return   = 0;
		memset(fi_.active, 0, sizeof(int)*(FI_MAX_FLT_LOC));
		fi_.watch_open_ptr = 0;
		memset(fi_.fds, 0, sizeof(fi_fltd_t*)*(FI_MAX_FLTD));
		fi_.fd             = -1;
		fi_.syscall        = -1;
		fi_.socketcall     = -1;
		fi_eprintf("fi_socket=%d\n", fi_socket) ;
		fi_.ncpu           = ncpu;
		if (fi_.ncpu != 1) {
			/* FIXME */
			fi_eprintf("Disabling fault injection.\nFault injection currently only for single CPU\n");
		}
	} else {
		fi_.pid = pid ;
		fi_init_netinfo();
	}
	if (fi_socket < 0) {
		fi_socket = fi_init_socket() ;
	}

	/* parse whatever fault descriptions are available right now */
	if (fi_io()) {
		fi_do();
	}
}
/***********************************************************************
 * fi_do
 ***********************************************************************
 * Input:     none
 * Output:    none
 * Effects:   Called from tracer's SIGIO handler.  Depending on the type of
 *            fault, the appropriate actions are taken to activate this fault.
 *            Appends fault description just activated to list of active faults
 *            (for permanent or duration faults).  Injects a transient fault at
 *            once.  SIGALRM is blocked other than in the central "waiting for
 *            child to stop" state, so unless this is changed, we should be ok
 *            when modifying fault list structures.
 * Called By: do_fault_injection in tracer.c, fi_init
 **********************************************************************/
void fi_do(void)
{
	while (fi_first != (fi_fltd_t *) 0) {
		fi_fltd_t *fltd ;

		/* take one fault description from list */
		fltd = fi_first;

		/* remove from list */
		fi_first = fltd->next;
		if (fltd->next == (fi_fltd_t *) 0) {
			fi_last = (fi_fltd_t *) 0;
		} else {
			fltd->next->prev = (fi_fltd_t *) 0;
		}

		/* do fault injection */
		activate_fltd(fltd) ;
	}
}
/***********************************************************************
 * fi_io
 ***********************************************************************
 * Input:	none
 * Output:	0: no new fault descriptions
 *		1: new fault descriptions arrived
 **********************************************************************/
int fi_io(void)
{
	char c;
	fi_fltd_t *fltd ;
	unsigned long num ;
	int ret;


	ret = read(fi_socket, &c, 1);
	if (ret < 0 && errno != EAGAIN) {
		fprintf(stderr, "read: %s.\n", strerror(errno));
		assert(0);
	}
	if (ret < 0) {
		/* non blocking io => nothing to read */
		assert(errno == EAGAIN);
		return 0;
	}
	if (ret == 0) {
		/* end of file */
		return 0;
	}

	fi_setsocket_sync(fi_socket);

	if (fi_node < 0) {
		ret = read_decnum(fi_socket, &num);
		fi_node = (int) num;
	}
	assert(0 <= fi_node);

	for (;;) {
		ret = parse_fltd(fi_socket, &fltd);

		if (ret == END_OF_FAULTLOAD) {
			break;
		}

		fltd->prev = fi_last;
		fltd->next = (fi_fltd_t *) 0;
		if (fltd->prev == (fi_fltd_t *) 0) {
			fi_first = fltd;
		} else {
			fltd->prev->next = fltd;
		}
		fi_last = fltd;
	}

	fi_setsocket_async(fi_socket);

	return fi_first != (fi_fltd_t *) 0;
}
