/* $Id: fi_print.c,v 1.1.1.1 2002/08/16 17:29:27 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/user.h>
#include <sys/types.h>

#include "fi_defs.h"

extern fi_flt_inj_t fi_; /* fi.c */
/*******************************************************************************
 * function prototypes
 *******************************************************************************
 * file internal functions 
 ******************************************************************************/
char *loc_to_str(int);
char *type_to_str(int);
char *reg_to_str(int);

void fi_dump_regs (struct user_regs_struct *);
void fi_lprintf(char *, ...);
void print_fd_hd_blocks(FILE*,fi_fltd_t*);
void print_fd_hd_fail(FILE*,fi_fltd_t*);
void print_fd_net(FILE *, fi_fltd_t*);
void print_fd_cpu_regs(FILE*,fi_fltd_t*);
void fi_print_afdl(fi_fltd_t *, char*);
/*******************************************************************************
 * exported functions
 ******************************************************************************/
void fi_dprintf(char *, ...);
void fi_eprintf(char *, ...);
void fi_print_all_dbg(fi_fltd_t *, char *);

/*******************************************************************************
 * functions to convert integer location/type/register identifiers to strings
 ******************************************************************************/
char *
loc_to_str(int loc){
	if (loc == FAULT_RAM)       return FAULT_RAM_STR;
	if (loc == FAULT_MMU)       return FAULT_MMU_STR;
	if (loc == FAULT_HD_BLOCKS) return FAULT_HD_BLOCKS_STR;
	if (loc == FAULT_HD_FAIL)   return FAULT_HD_FAIL_STR;
	if (loc == FAULT_CPU_REGS)  return FAULT_CPU_REGS_STR;
	if (loc == FAULT_NET_SEND)  return FAULT_NET_SEND_STR;
	if (loc == FAULT_NET_RECV)  return FAULT_NET_RECV_STR;
	if (loc == FAULT_POWER)     return FAULT_POWER_STR;
	if (loc == FAULT_DURATION_END) return FAULT_DURATION_END_STR;
	/* should not be reached */
   return (char*) NULL;
}
/*******************************************************************************
 ******************************************************************************/
char *
type_to_str(int type){
	if (type == FAULT_PERMANENT) return FAULT_PERMANENT_STR;
	if (type == FAULT_TRANSIENT) return FAULT_TRANSIENT_STR;
	if (type == FAULT_DURATION) return FAULT_DURATION_STR;
	if (type == FAULT_DURATION_END) return FAULT_DURATION_END_STR;
	/* should not be reached */
   return (char*) NULL;
}
/*******************************************************************************
 ******************************************************************************/
char *
reg_to_str(int reg){
	if (reg == CPU_REG_EAX) return CPU_REG_EAX_STR;
	if (reg == CPU_REG_EBX) return CPU_REG_EBX_STR;
	if (reg == CPU_REG_ECX) return CPU_REG_ECX_STR;
	if (reg == CPU_REG_EDX) return CPU_REG_EDX_STR;
	if (reg == CPU_REG_ESI) return CPU_REG_ESI_STR;
	if (reg == CPU_REG_EDI) return CPU_REG_EDI_STR;
	if (reg == CPU_REG_EBP) return CPU_REG_EBP_STR;
	if (reg == CPU_REG_ESP) return CPU_REG_ESP_STR;
	if (reg == CPU_REG_EIP) return CPU_REG_EIP_STR;
	if (reg == CPU_REG_DS) return CPU_REG_DS_STR;
	if (reg == CPU_REG_CS) return CPU_REG_CS_STR;
	if (reg == CPU_REG_SS) return CPU_REG_SS_STR;
	if (reg == CPU_REG_ES) return CPU_REG_ES_STR;
	if (reg == CPU_REG_FS) return CPU_REG_FS_STR;
	if (reg == CPU_REG_GS) return CPU_REG_GS_STR;
	/* should not be reached */
   return (char*) NULL;
}
/*******************************************************************************
 * adapted version of fprintf for error output to file/stderr
 ******************************************************************************/
void
fi_eprintf(char *fmt, ...)
{
	va_list ap;
	FILE *file;
	va_start(ap, fmt);

	if (fmt == NULL){
		unlink(FI_ERR_FILE);
		return;
	}

#if FI_DBG
	file=stderr ;
#else
	file = fopen(FI_ERR_FILE, "a");
	assert (file != NULL );
#endif
	vfprintf(file, fmt, ap);
#if FI_DBG
	fflush(file);
#else
	fclose(file);
#endif
	va_end(ap);
}
/*******************************************************************************
 * adapted version of fprintf for debugging and logging output
 ******************************************************************************/
void
fi_dprintf(char *fmt, ...)
{
#if FI_DBG
	va_list ap;
	FILE *file;

	if (fmt == NULL){
		unlink(FI_DBG_FILE);
		return;
	}

	file = fopen(FI_DBG_FILE, "a");
	assert (file != NULL );
	va_start(ap, fmt);
	vfprintf(file, fmt, ap);
	fclose(file);
	va_end(ap);
#endif
}
/*******************************************************************************
 ******************************************************************************/
void
fi_lprintf(char *fmt, ...)
{
#if FI_LOG
	va_list ap;
	FILE *file;

	assert(FI_LOG_FILE != 0 );
	if (fmt == NULL){
		unlink(FI_LOG_FILE);
		return;
	}

	file = fopen(FI_LOG_FILE, "a");
	assert (file != NULL );
	va_start(ap, fmt);
	vfprintf(file, fmt, ap);
	fclose(file);
	va_end(ap);
#endif
}
/*******************************************************************************
 ******************************************************************************/
void
fi_dump_regs(struct user_regs_struct *regs)
{
	assert(regs != 0);
	fi_dprintf("\n   ebx=%08lx    ecx=%08lx   edx=%08lx\n",
			(*regs).ebx, (*regs).ecx, (*regs).edx);

	fi_dprintf("   esi=%08lx    edi=%08lx   ebp=%08lx\n",
			regs->esi, regs->edi, regs->ebp);

	fi_dprintf("   eax=%08lx             orig_eax=%08lx\n",
			regs->eax, regs->orig_eax);

	fi_dprintf("   eip=%08lx eflags=%08lx   esp=%08lx\n",
			regs->eip, regs->eflags, regs->esp);

	fi_dprintf("   ds=%08lx   es=%08lx\n", regs->xds, regs->xes);

	fi_dprintf("   fs=%08lx   gs=%08lx\n", regs->xfs, regs->xgs);

	fi_dprintf("   cs=%08lx   ss=%08lx\n", regs->xcs, regs->xss);
}
/*******************************************************************************
 ******************************************************************************/
void
print_fd_hd_blocks(
	FILE *file,
	fi_fltd_t* fltd)
{
	fprintf(file, "   %s # device\n", (fltd->fltd).hd.device);
	fprintf(file, "   %lld # block\n", (fltd->fltd).hd.block);
	fprintf(file, "   %lld # range\n", (fltd->fltd).hd.range);
}
/*******************************************************************************
 ******************************************************************************/
void
print_fd_hd_fail(
	FILE *file,
	fi_fltd_t* fltd)
{
	fprintf(file, "   %s # device \n", (fltd->fltd).hd.device);
}
/*******************************************************************************
 ******************************************************************************/
void
print_fd_net(
	FILE *file,
	fi_fltd_t* fltd)
{
	fprintf(file, "   %s # device \n",(fltd->fltd).net.device);
	fprintf(file, "   %d # fd (may be uninitialized)\n",(fltd->fltd).net.fd);
	fprintf(file, "   %d # =0x%x port\n",
			(fltd->fltd).net.port,(fltd->fltd).net.port);
}
/*******************************************************************************
 ******************************************************************************/
void
print_fd_ram(
	FILE *file,
	fi_fltd_t* fltd)
{
	fprintf(file, "   %ld # = 0x%lx address \n",
			(fltd->fltd).ram.addr, (fltd->fltd).ram.addr);
	fprintf(file, "   %d # bit \n", (fltd->fltd).ram.bit);
}
/*******************************************************************************
 ******************************************************************************/
void
print_fd_cpu_regs(
	FILE *file,
	fi_fltd_t* fltd)
{
	fprintf(file, "   %s # register \n",reg_to_str((fltd->fltd).cpu.reg));
	fprintf(file, "   %d # bit \n", (fltd->fltd).cpu.bit);
}
/*******************************************************************************
 * fi_print_afdl
 * Input: fdl_start = pointer to start of list to print
 *        fname = name of file to print to
 * Output: dump of fault description list into file
 *
 ******************************************************************************/
void
fi_print_afdl(
	fi_fltd_t *fdl_start,
	char *fname)
{
	fi_fltd_t *fdl_current;
	FILE *file;

	assert(fname != 0);


	file = fopen(fname, "a");

	assert(file!=0);

	if (fdl_start == NULL){
		fprintf(file, "No fault descriptions available!\n");
	}else{
		fdl_current = fdl_start;
		while(fdl_current != NULL) {
			fprintf(file, "%s # id %d, 0x%p next=0x%p\n",
				loc_to_str(fdl_current->location),
				fdl_current->id, fdl_current, fdl_current->next);
			fprintf(file, "   %s\n", type_to_str(fdl_current->type));
			fprintf(file, "   %ld # activation time\n",
				fdl_current->atime);
			switch (fdl_current->location) {
				case FAULT_HD_BLOCKS:
					print_fd_hd_blocks(file,fdl_current);
					break;
				case FAULT_HD_FAIL:
					print_fd_hd_fail(file,fdl_current);
					break;
				case FAULT_RAM:
					print_fd_ram(file,fdl_current);
					break;
				case FAULT_CPU_REGS:
					print_fd_cpu_regs(file,fdl_current);
					break;
				case FAULT_NET_SEND:
				case FAULT_NET_RECV:
					print_fd_net(file,fdl_current);
					break;
				case FAULT_DURATION_END:
					break;
				default:
					fprintf(file,
					"   Fault type not supported yet.\n");
					break;

			}
			fdl_current = fdl_current->next;
		}
	}

	fclose(file);
}
/*******************************************************************************
 ******************************************************************************/
void
fi_print_all_dbg(
	fi_fltd_t *fdl_start,
	char *fname)
{
	int i;
	FILE *file;

	assert(fname != 0);


	file = fopen(fname, "w");
	fclose(file);
  	fi_print_afdl(fdl_start, fname);

	file = fopen(fname, "a");
	assert(file != NULL);

	fprintf(file, "pid: %d\n", fi_.pid);
	fprintf(file, "watch_enter/-_return: %lx/%lx\n",
			fi_.watch_enter, fi_.watch_return);
	fprintf(file, "num_fltd %d\n", fi_.num_fltd);
	fprintf(file,"watch_open_ptr: %lx\n",(unsigned long)fi_.watch_open_ptr);

	/* print file descriptor watch stuff */
	for (i=0; i<FI_MAX_FLTD; i++){
		int j = 0;
		fprintf(file, "fds[%2d]= ",i);
		for (j=0; j<FI_MAX_FLTD; j++){
			fprintf(file, "\ta[%d]=0x%p, ",j, (fi_.fds[i]).a[j]);
			if((fi_.fds[i]).a[j] == 0) break ;
		}
		fprintf(file, "\n");
	}
	/*
	fprintf(file, "fi_fdl_start=%lx\n",(unsigned long)fi_fdl_start);
	fprintf(file, "fi_fdl_anext=%lx\n",(unsigned long)fi_fdl_anext);
	*/

	/* print netinfo */
	for (i=0; i<MAX_ETH_CARDS; i++){
		fprintf(file, "netinfo[%2d]: 0x%x <=> %d\n", i, fi_.netinfo[i].port,
				fi_.netinfo[i].fd) ;
	}
	fclose(file);
}
