/*
 * $Id: main.c,v 1.5 2003/02/25 23:33:30 mbasrai Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/wait.h>
#include <asm/unistd.h>

//#include "lotol.h"
#include "lotol_fdesc.h"
//#include "fi_interface.h"
#include "lilo.h"
#include "config.h"

char *progname;
int opt_p = 0;
int opt_c = 1;	/* 1 CPU default */
int opt_mp = 0;	/* default: uniprocessor */
int pids [MAX_CPU] = { -1, -1, /*...*/ -1 };

//used for debugging STK
#if 0
#define _lotol_syscall_return(type, res) \
do { \
	if ((unsigned long)(res) >= (unsigned long)(-125)) { \
		errno = -(res); \
		res = -1; \
	} \
	return (type) (res); \
} while (0)


#define _lotol_syscall1(type,name,type1,arg1) \
type lotol_##name(type1 arg1) \
{ \
	long __res; \
	__asm__ volatile ("int $0x80" \
		: "=a" (__res) \
		: "0" (0),"b" ((long)(arg1)), \
		  "D" (__NR_##name << 16)); \
	_lotol_syscall_return(type,__res); \
}

#define __NR_syncprint 237
static inline _lotol_syscall1(int,syncprint,char *,ubuf);
#endif

/* Initialize hardware */
static void
hwinitialize(void)
{
	umcore_init(opt_c);
	ummem_init();
	umrtc_init();
	umkeyboard_init();
	umvideo_init();
	umfloppy_init();
	umide_init();
	umnet_init();
	if (opt_mp) {
		umapic_init(opt_c);
	}
}

/* Shutdown hardware */
static void
hwshutdown(void)
{
	umnet_shutdown();
	umide_shutdown();
	umfloppy_shutdown();
	umvideo_shutdown();
	umkeyboard_shutdown();
	umrtc_shutdown();
	ummem_shutdown();
	umcore_shutdown(opt_c);
	if (opt_mp) {
		umapic_shutdown(opt_c);
	}
}

/* Initialize bios variables */
static void
biosinitialize(void)
{
	/* no drive_config in RTC */
	*(unsigned char *) (VA_RTC + 18) = 0;

	umsmp_init(opt_c);
}

static void
do_cpu(int nr)
{
	void (*entry)(int);
	//int ret;
	
	ummem_CPU_init();
	umcore_CPU_init(nr, opt_c);
	umapic_CPU_init(nr, opt_c);
	umrtc_CPU_init();
	umvideo_CPU_init();
	umrtc_CPU_init();
	
	printf("nr == %d\n",nr);
	//ret = ptrace(PTRACE_TRACEME, 0, 0, 0);
	//assert(0 <= ret);

	//ret = kill(getpid(), SIGTRAP);
	//assert(0 <= ret);
	
	ptrace(25,0,0,0);

	if (nr == 0) {
		char **argv;

		/*
		 * first cpu initializes node
		 */
		/* initialize bios */
		biosinitialize();

		/* read lilo.conf */
		argv = lilo_conf_read();

		/* load image to high memory */
		//fprintf(stderr, "About to load kernel...\n");
		entry = do_load(argv);

	} else {
		/*
		 * other cpus wait for IPI
		 */
		entry = do_wait(nr);
	}
	//fprintf(stderr, "About to call *entry(nr)\n");

	(*entry)(nr);
}


static void
usage(int retval)
{
	fprintf(stderr, "Usage: %s [-c <n>] [-p] <system>\n", progname);
	fprintf(stderr, "\t-p: protect kernel memory\n");
	fprintf(stderr, "\t-c <n>: run n cpus, 1 <= n <= %d\n", MAX_CPU);
	exit(retval);
}

int
main(int argc, char **argv)
{
	extern int tracer(int ncpu);
	int c;
	int i;
	int ret;

	/*
	 * get program name
	 */
	progname = *argv;


	printf("lilo main.c\n");
	/*
	 * get options
	 */
	while ((c = getopt(argc, argv, "pc:")) != -1) {
		switch (c) {
		case 'p':
			opt_p = 1;
			break;
		case 'c':
			opt_c = atoi(optarg);
			if (opt_c < 1 || MAX_CPU < opt_c)
				usage(1);
			opt_mp = 1;
			break;
		default:
			usage(1);
		}
	}
	argc -= optind;
	argv += optind;

	/* read hardware configuration */
	printf("config_read\n");
	config_read();

	/* 
	 * Put us in our own process group, which will receive
	 * SIGIO and SIGURG for events on file descriptor fd.
	 *
	 * XXX Handling of these signals should be serialized by
	 * SMP code the kernel.
	 */ 
	ret = setpgrp();
	assert(0 <= ret);

	//printf("fi_init\n");
	//fi_init(0, opt_c); /* initialize faultinjection */

	/* setup hardware */
	printf("hwinit\n");
	hwinitialize();

	do {
		/*
		 * fork
		 * - children will run image
		 * - parent will remap system calls
		 */
		for (i = 0; i < opt_c; i++) {
			//pids[i] = fork();
			//switch (pids[i]) {
			//case -1:
				//fprintf(stderr, "%s: fork: %s.\n", progname,
			//strerror(errno));
				//exit(1);
			//case 0:
				/* child */
			do_cpu(i);
				/*NOTREACHED*/
			assert(0);
			
			//default:
				/* parent */
				//break;
			//}
		}

		/* FIXME: should use pids[] array! */
		//fi_init(pids[0], opt_c); /* re-initialize faultinjection */
		pids[i] = -1;

#ifdef SMP	/* FIXME HSHOEXER */
		/* Create APIC */
		umapic_fork(opt_c);
#endif /* SMP */

		/* Run Tracer */
		ret = tracer(opt_c);

	} while (ret == 0);

	hwshutdown();

	//fi_exit() ; /* close faultinjection socket to frontend */

	return (0);
}

/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
