/*
 * $Id: umeth.c,v 1.5 2003/02/20 23:58:29 mbasrai Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <net/if.h>

#include "external_constants.h"
#include "lotol_fdesc.h"

#include "lilo.h"

/*
 * Atmost one of HOST_TUNTAP_ORIGINAL and HOST_TUNTAP_MODIFIED may be defined
 * based on whether we want to use the original or modified host TUN/TAP driver
 */
/* #define HOST_TUNTAP_ORIGINAL */
#define HOST_TUNTAP_MODIFIED

#ifdef HOST_TUNTAP_ORIGINAL
#ifdef HOST_TUNTAP_MODIFIED
#error Cannot define both HOST_TUNTAP_ORIGINAL and HOST_TUNTAP_MODIFIED
#endif
#endif

/*
 * When either HOST_TUNTAP_ORIGINAL or HOST_TUNTAP_MODIFIED is defined,
 * HOST_TUNTAP_COMMON must be defined
 */
#ifdef HOST_TUNTAP_ORIGINAL
#define HOST_TUNTAP_COMMON
#endif
#ifdef HOST_TUNTAP_MODIFIED
#define HOST_TUNTAP_COMMON
#endif

void
umnet_init(void)
{
	int fd;
#ifndef HOST_TUNTAP_COMMON
	int eth;
#endif
	int ret;
	int i;
	int flags;
	unsigned int myself;
#ifndef HOST_TUNTAP_COMMON
	char name[] = "eth.";
#endif
	char line[80];
	char *sret;
#ifndef HOST_TUNTAP_COMMON
	unsigned short ipport;
	struct sockaddr_in addr;
#endif
	FILE *sfd;
#ifdef HOST_TUNTAP_ORIGINAL
	char nibble1, nibble2, digit1, digit2;
	struct ifreq ifr;
	struct ifreq ifr1;
#endif
#ifdef HOST_TUNTAP_MODIFIED
	struct ifreq ifr;
	unsigned char mac_address[6];
#endif
#ifndef HOST_TUNTAP_COMMON
	for (eth = 0; eth < MAX_ETH_CARDS; eth++) {
		sprintf(name, "eth%d", eth);

		fd = open(name, O_RDONLY, 0);
		if (fd < 0 && errno != ENOENT) {
			fprintf(stderr, "%s: %s: %s.\n", progname,
					name, strerror(errno));
			exit(1);
		}
		if (fd < 0) {
			/* no such interface */
			continue;
		}
		assert(0 <= fd && fd < FD_LOTOL_MIN);

		ret = dup2(fd, FD_ETH_CONF(eth));
		assert(ret == FD_ETH_CONF(eth));

		ret = close(fd);
		assert(0 <= ret);

		fd = FD_ETH_CONF(eth);

		/* open an I/O-stream, much easier to handle... */
		sfd = fdopen(fd, "r");
		assert(NULL != sfd);

		sret = fgets(line, sizeof(line) - 1, sfd);
		assert(sret == line);

		sscanf(line, "%x", &myself);	

		/* Skip over to "our" line */
		for (i = 0; i <= myself; i++) {
			sret = fgets(line, sizeof(line) - 1, sfd);
			assert(sret == line);
		}

		/* reset descriptor */
		ret = lseek(fd, 0, 0);
		assert(ret == 0);

		sscanf(line, "%*x:%*x:%*x:%*x:%*x:%*x %*x:%hx\n", &ipport);

		/* create socket */
		fd = socket(PF_INET, SOCK_DGRAM, 0);
		assert(0 <= fd && fd < FD_LOTOL_MIN);	

		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons(ipport);
		addr.sin_addr.s_addr = INADDR_ANY;

		ret = dup2(fd, FD_ETH(eth));
		assert(ret == FD_ETH(eth));

		ret = close(fd);
		assert(0 <= ret);

		fd = FD_ETH(eth);

		dfprintf(stderr, "%s: binding %s(%d) [0x%04x].\n",
				progname, name, fd, ipport);
		ret = bind(fd, (struct sockaddr *)&addr, sizeof(addr));
		if (ret < 0) {
			fprintf(stderr, "%s: Can't bind %s(%d) [0x%04x]: %s.\n",
					progname, name, fd, ipport,
					strerror(errno));
			exit(1);
		}
		assert(0 <= ret);

		flags = fcntl(fd, F_GETFL, 0);
		flags |= O_ASYNC;
		flags |= O_NONBLOCK;
		ret = fcntl(fd, F_SETFL, flags);
		assert(0 <= ret);

		ret = fcntl(fd, F_SETOWN, -1 * getpgrp());
		assert(0 <= ret);
	}
#else
#ifdef HOST_TUNTAP_MODIFIED
	/* Read the MAC address from the tuntap file */
	fd = open("tuntap", O_RDONLY, 0);
	if (fd < 0) {
		fprintf(stderr, "%s: %s: %s.\n", progname,
			"tuntap", strerror(errno));
		exit(1);
	}
	assert(0 <= fd && fd < FD_LOTOL_MIN);
	
	ret = dup2(fd, FD_TUNTAP_CONF);
	assert(ret == FD_TUNTAP_CONF);
	
	ret = close(fd);
	assert(0 <= ret);
	
	fd = FD_TUNTAP_CONF;

	/* open an I/O-stream, much easier to handle... */
	sfd = fdopen(fd, "r");
	assert(NULL != sfd);
	
	sret = fgets(line, sizeof(line) - 1, sfd);
	assert(sret == line);
	
	sscanf(line, "%x", &myself);	
	
	/* Skip over to "our" line */
	for (i = 0; i <= myself; i++) {
		sret = fgets(line, sizeof(line) - 1, sfd);
		assert(sret == line);
	}
	
	ret = fflush(sfd);
	assert(ret == 0);
	
	sscanf(line, "%hx:%hx:%hx:%hx:%hx:%hx %*x:%*x\n", &mac_address[0],
	       &mac_address[1], &mac_address[2], &mac_address[3],
	       &mac_address[4], &mac_address[5]);
#endif
	/* Setup the tun_tap device */
	fd = open("/dev/net/tun", O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "%s: /dev/net/tun: %s.\n", progname,
			strerror(errno));
		exit(1);
	}
	assert(0 <= fd && fd < FD_LOTOL_MIN);
	
	errno = 0;
	ret = dup2(fd, (FD_TUNTAP));
	assert(ret == FD_TUNTAP);

	fprintf(stderr, "Setting network FD to %d\n", FD_TUNTAP);
#define __NR_ztrace 233
#define ZTRACE_SET_OPT 22
#define ZTRACE_MARKER 32
#define REPLAY_OPT_VALUE (1 << 16)
#define REPLAY_OPT_NETWORKFD  (REPLAY_OPT_VALUE | 3)

	if(syscall(__NR_ztrace, ZTRACE_SET_OPT, getpid(), 
			REPLAY_OPT_NETWORKFD, FD_TUNTAP)) {
		perror("Setting network fd");
	}
	
	ret = close(fd);
	assert(0 <= ret);
	
	fd = FD_TUNTAP;
#ifdef HOST_TUNTAP_MODIFIED
	if(ioctl(fd, TUNSETMACADD, (void *) mac_address) < 0) {
		perror("TUNSETMACADD");
		exit(1);
	}
#endif
	memset(&ifr, 0, sizeof(ifr));
	
	ifr.ifr_flags = (IFF_TAP | IFF_NO_PI);
	strncpy(ifr.ifr_name, "", sizeof(ifr.ifr_name) - 1);

	if(ioctl(fd, TUNSETIFF, (void *) &ifr) < 0) {
		perror("TUNSETIFF");
		exit(1);
	}
   
	if(ioctl(fd, TUNSETNOCSUM, 1) < 0) {
		perror("TUNSETNOCSUM");
		exit(1);
	}
	/*
	if(ioctl(fd, TUNSETPERSIST, 1) < 0) {
	        perror("TUNSETPERSIST");
		exit(1);
	}
	*/
	if(ioctl(fd, TUNSETOWNER, getuid()) < 0) {
	        perror("TUNSETOWNER");
	        exit(1);
	}

	flags = fcntl(fd, F_GETFL, 0);
	flags |= O_ASYNC;
	flags |= O_NONBLOCK;
	ret = fcntl(fd, F_SETFL, flags);
	assert(0 <= ret);
	
	ret = fcntl(fd, F_SETOWN, -1 * getpgrp());
	assert(0 <= ret);
	
	/* Write interface name to tuntapname file */
	fd = creat("tuntapname", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	assert(0 <= fd);

	ret = write(fd, ifr.ifr_name, strlen(ifr.ifr_name));
	assert(ret == strlen(ifr.ifr_name));

	ret = close(fd);
	assert( ret == 0 );
#ifdef HOST_TUNTAP_ORIGINAL
	/* Create a channel to the NET kernel */
	if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	/* Get the hw address of the tap interface */
	bzero(&ifr1, sizeof(ifr1));
	strcpy(ifr1.ifr_name, ifr.ifr_name);
	if (ioctl(fd, SIOCGIFHWADDR, &ifr1) < 0) {
		perror ("ioctl(SIOCGIFHWADDR)");
		exit (1);
	}
	
	ret = close(fd);
	assert(0 <= ret);

	/* Write it to the tuntap file */
	fd = open("tuntap", O_RDWR, 0);
	if (fd < 0) {
		fprintf(stderr, "%s: %s: %s.\n", progname,
			"tuntap", strerror(errno));
		exit(1);
	}
	assert(0 <= fd && fd < FD_LOTOL_MIN);
	
	ret = dup2(fd, FD_TUNTAP_CONF);
	assert(ret == FD_TUNTAP_CONF);
	
	ret = close(fd);
	assert(0 <= ret);
	
	fd = FD_TUNTAP_CONF;

	/* open an I/O-stream, much easier to handle... */
	sfd = fdopen(fd, "r+");
	assert(NULL != sfd);
	
	sret = fgets(line, sizeof(line) - 1, sfd);
	assert(sret == line);
	
	sscanf(line, "%x", &myself);	
	
	/* Skip over to begining of "our" line */
	for (i = 0; i < myself; i++) {
		sret = fgets(line, sizeof(line) - 1, sfd);
		assert(sret == line);
	}
	for (i = 0; i < 6; i++) {
		nibble1 = (ifr1.ifr_hwaddr.sa_data[i] & 0xf0) >> 4;
		nibble2 = ifr1.ifr_hwaddr.sa_data[i] & 0x0f;
		if ( nibble1 < 10 )
			digit1 = nibble1 + '0';
		else
			digit1 = nibble1 - 10 + 'a';
		if ( nibble2 < 10 )
			digit2 = nibble2 + '0';
		else
			digit2 = nibble2 - 10 + 'a';
		ret = fputc(digit1, sfd);
		assert(ret == digit1);
		ret = fputc(digit2, sfd);
		assert(ret == digit2);
		fputc( ((i == 5) ? ' ' : ':') , sfd);
	}
	
	ret = fflush(sfd);
	assert(ret == 0);
#endif
#endif /* HOST_TUNTAP_COMMON */
}

/* Try to shutdown ethernet */
void
umnet_shutdown(void)
{
	int eth;

	for (eth = 0; eth < MAX_ETH_CARDS; eth++) {
		(void) close(FD_ETH(eth));
		(void) close(FD_ETH_CONF(eth));
	}
#ifdef HOST_TUNTAP_COMMON
	(void) close(FD_TUNTAP);
	(void) close(FD_TUNTAP_CONF);
#endif
}
