/*
 * $Id: umvideo.c,v 1.4 2003/02/27 18:56:46 dunlapg Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include "lotol_fdesc.h"
#include "lilo.h"
#include "config.h"

#define PALETTE_SIZE 256
#define MAX_VIDEO_SIZE 1
#define VIDEO_SIZE 6*65536


#define STDERR_MSG(msg) fprintf(stderr, "%s:%d: %s\n", \
		__FILE__, __LINE__, msg)

typedef struct {
	pid_t	pid;
	char	*vid_map;
	char	*pal_map;
	int	width;
	int	height;
	int	depth;
	int	scale;
	int	vmap_size;
	char	*layers[2];
        int     (*update_palette)(void);
        void    (*do_video)(int sat);
        char    pal_old[PALETTE_SIZE * 3];
        int     sl_jmp;
} video_struct;

static video_struct *vsd;

static int vt_width;
static int vt_height;
static int vt_rows;
static int vt_sum;

static char	*vt_packet;
static int	vt_size;

/* mem waste ;-) */
static char	init_packet[7];

static void
send_video_packet(char *packet)
{
	// UMLfront is no longer used. The packets are directly written to 
	// stdout in the guest video driver -MAB 
	return;
#if 0
	int ret;
	int len = 0;

	do {
		ret = write(FD_DISPLAY, packet + len, vt_size - len);
		if (ret < 0) {
			perror("umvideo write");
			ret = 0;
		}
		assert(0 <= ret);
		len += ret;
	} while (len != vt_size);
#endif
}


static int
dummy_update_palette(void)
{
        return 0;
}

static int
update_palette_8bpp(void)
{
        int send_all_tile = 0;
        int i;
        int pidx = 0;

        vt_packet[0] = 7;

        for(i = 0; i < PALETTE_SIZE; i++) {
                if(memcmp(vsd->pal_old + i * 3, vsd->pal_map + i * 3, 3)) {
                        *(vt_packet + 3 + pidx * 4) = i;
                        memcpy(vt_packet + 3 + pidx * 4 + 1,
                               vsd->pal_map + i * 3, 3);
                        pidx++;
                }
                if(pidx == 32 || (i == PALETTE_SIZE - 1 && pidx)) {
                        *((short *) (vt_packet + 1)) = pidx;
			send_video_packet(vt_packet);
                        send_all_tile = 1;
                        pidx = 0;
                }
        }

        memcpy(vsd->pal_old, vsd->pal_map, PALETTE_SIZE * 3);

        return send_all_tile;
}

static void
do_video_generic(int sat)
{
        static int ul_idx = 0;
        int     i;
        int     vt_idx = 0;
        char    *ulp = vsd->layers[ul_idx];
        char    *olp = vsd->layers[1 - ul_idx];
	int	send;
	int	vd_ssize = vsd->width * vsd->scale;
	int	vt_ssize = vt_width * vsd->scale;
	
        vt_packet[0] = 6;
        memcpy(ulp, vsd->vid_map, vsd->vmap_size);

        do {
		send = sat;
                for (i = 0; i < 16; i++) {
                        if (memcmp(ulp + i * vd_ssize, olp + i * vd_ssize,
                                   vt_ssize)) {
				send = 1;
				break;
			}
		}
		if (send) {
			*((short *) (vt_packet + 1)) = vt_idx;
			for (i = 0; i < 16; i++) {
				memcpy(vt_packet + 3 + i * vt_ssize,
				       ulp + i * vd_ssize, vt_ssize);
			}
			send_video_packet(vt_packet);
		}

                vt_idx++;

                if(vt_idx % vt_rows) {
                        ulp += vt_ssize;
                        olp += vt_ssize;
                }
                else {
                        ulp += vsd->sl_jmp;
                        olp += vsd->sl_jmp;
                }
        } while(vt_idx < vt_sum);

        ul_idx = 1 - ul_idx;
}


static int
setup_video_sim(int mode)
{
        vsd->update_palette = dummy_update_palette;
        vsd->do_video = do_video_generic;

	init_packet[0] = 0;
	*((unsigned short *) (init_packet + 1)) = mode;
	vt_size = 7;
	
	switch(mode) {
	/*
	 * 8 bit
	 */
	case 769:
		vsd->width	= 640;
		vsd->height	= 480;
		vsd->depth	= 8;
		vt_width	= 8;
		vt_height	= 16;
                vsd->update_palette = update_palette_8bpp;
		break;
	case 771:
		vsd->width	= 800;
		vsd->height	= 600;
		vsd->depth	= 8;
                vt_width        = 8;
		vt_height	= 16;
		vsd->update_palette = update_palette_8bpp;
		break;
	case 773:
		vsd->width	= 1024;
		vsd->height	= 768;
		vsd->depth	= 8;
		vt_width	= 8;
		vt_height	= 16;
                vsd->update_palette = update_palette_8bpp;
		break;
	case 775:
		vsd->width	= 1280;
		vsd->height	= 1024;
		vsd->depth	= 8;
		vt_width	= 8;
		vt_height	= 16;
                vsd->update_palette = update_palette_8bpp;
		break;
	/*
	 * 15 bit
	 */
	case 784:
		vsd->width      = 640;
		vsd->height     = 480;
		vsd->depth      = 16;
		vt_width        = 8;
		vt_height       = 16;
		break;
	case 787:
		vsd->width      = 800;
	        vsd->height     = 600;
	        vsd->depth      = 16;
	        vt_width        = 8;
	        vt_height       = 16;
		break;
	case 790:
		vsd->width      = 1024;
		vsd->height     = 768;
		vsd->depth      = 16;
		vt_width        = 8;
		vt_height       = 16;
		break;
	case 793:
		vsd->width      = 1280;
		vsd->height     = 1024;
		vsd->depth      = 16;
		vt_width        = 8;
	        vt_height       = 16;
		break;
	/*
	 * 16 bit
	 */
	case 785:
                vsd->width      = 640;
                vsd->height     = 480;
                vsd->depth      = 16;
		vt_width	= 8;
		vt_height	= 16;
		break;
	case 788:
		vsd->width      = 800;
                vsd->height     = 600;
           	vsd->depth      = 16;
                vt_width        = 8;
		vt_height	= 16;
		break;
	case 791:
		vsd->width      = 1024;
		vsd->height     = 768;
		vsd->depth      = 16;
		vt_width        = 8;
	        vt_height       = 16;
		break;
	case 794:
		vsd->width      = 1280;
                vsd->height     = 1024;
		vsd->depth      = 16;
		vt_width        = 8;
		vt_height       = 16;
		break;
	case 786:
		vsd->width	= 640;
		vsd->height	= 480;
		vsd->depth	= 32;
		vt_width	= 8;
		vt_height	= 16;
		break;
	case 789:
		vsd->width      = 800;
		vsd->height     = 600;
		vsd->depth      = 32;
		vt_width        = 8;
		vt_height       = 16;
		break;
	case 792:
		vsd->width      = 1024;
		vsd->height     = 768;
		vsd->depth      = 32;
		vt_width        = 8;
		vt_height       = 16;
		break;
	case 795:
		vsd->width      = 1280;
		vsd->height     = 1024;
		vsd->depth      = 16;
		vt_width        = 8;
		vt_height       = 16;
		break;
	default:
		STDERR_MSG("invalid mode: oops: missing fallback imp");
		return 1;
	}

	/*
	 * send init packet --> move to video.c
	 */
	init_packet[3] = vt_width;
	init_packet[4] = vt_height;
	send_video_packet(init_packet);

	vsd->scale = vsd->depth / 8;
	vsd->sl_jmp = vt_width * vsd->scale +
		     (vt_height - 1) * vsd->width * vsd->scale;
	
	
	/* 24 vs. 32 bit --> depth? */
	vsd->vmap_size = vsd->width * vsd->height * vsd->scale;;
        vt_size = vt_width * vt_height * vsd->scale + 3;
        vt_rows = vsd->width / vt_width;
        vt_sum  = vt_rows * (vsd->height / vt_height);

	fprintf(stderr, "%d\n", vt_sum);

	if (!(vsd->layers[0] = (char *) malloc(vsd->vmap_size * sizeof(char)))
	||  !(vsd->layers[1] = (char *) malloc(vsd->vmap_size * sizeof(char))))
        {
		STDERR_MSG("layer buffer allocation failure");
		return 1;
	}

	if (!(vt_packet = (char *) malloc(vt_size * sizeof(char)))) {
		STDERR_MSG("buffer for dist sim allocation failure");
		return 1;
	}

        return 0;
}

#if 0

static int
sync_with_frontend(int mode)
{
	int ret;
	int len = 0;
	
	init_packet[0] = 0;
	*((unsigned short *) (init_packet + 1)) = mode;
	
	send_video_packet(init_packet);

	do {
		ret = read(FD_DISPLAY, init_packet + len, 7 - len);
		if(ret < 0) {
			perror("umvideo read");
			ret = 0;
		}
		assert(0 <= ret);
		len += ret;
	} while(len != 7);
	
	if(init_packet[0] != 0) {
		STDERR_MSG("warning: invalid init packet --> frontend panic");
		exit(1);
	}

	switch(*((unsigned short *) (init_packet + 1))) {
	case 0:
		/* force console
		 */
		return 1;
		break;
	case 1:
		/* valid video mode */
		break;
	default:
		STDERR_MSG("warning: invalid sync data");
		exit(1);
	}
	
	return 0;
}

#endif

void
umvideo_CPU_init(void)
{
	int ret;

	ret = (int) mmap((char *) VA_VIDEO_CHIP, SZ_VIDEO_CHIP,
			PROT_READ | PROT_WRITE,
			MAP_SHARED | MAP_FIXED, FD_VIDEO_CHIP, 0);
	assert(ret == VA_VIDEO_CHIP);

	/* FIXME cgtruong */
	ret = (int) mmap((char *) MEM_BASE + 0xa0000, 128*1024 * 3,
			PROT_READ | PROT_WRITE,
			MAP_SHARED | MAP_FIXED, FD_VIDEO_MEMORY, 0);
	assert(ret == MEM_BASE + 0xa0000);
}

void
umvideo_init(void)
{
	int i;
        int sat;
	char tmpfile[128];
	int ret;
	int fd;
	//struct sockaddr_in addr;

	/*
	 * "create" video chip
	 */
	sprintf(tmpfile, "/tmp/umlinux/video_chip%d", getpid());
	fd = open(tmpfile, O_EXCL | O_CREAT | O_RDWR);
	if(fd <= 0) {
		sprintf(tmpfile, "/tmp/video_chip%d", getpid());
		fd = open(tmpfile, O_EXCL | O_CREAT | O_RDWR);
	}
	assert(0 <= fd);

#if 0
	ret = unlink(tmpfile);
	assert(0 <= ret);
#endif

	{
		static char video_chip[SZ_VIDEO_CHIP];

		ret = write(fd, video_chip, sizeof(video_chip));
		assert(ret == sizeof(video_chip));
	}

	ret = dup2(fd, FD_VIDEO_CHIP);
	assert(ret == FD_VIDEO_CHIP);

	ret = close(fd);
	assert(0 <= ret);

	/*
	 * "create" video memory
	 */
	sprintf(tmpfile, "/tmp/umlinux/video_memory%d", getpid());
	fd = open(tmpfile, O_EXCL | O_CREAT | O_RDWR);
	if(fd <= 0) {
		sprintf(tmpfile, "/tmp/video_memory%d", getpid());
		fd = open(tmpfile, O_EXCL | O_CREAT | O_RDWR);
	}
	assert(0 <= fd);

#if 0
	ret = unlink(tmpfile);
	assert(0 <= ret);
#endif

	for (i = 0; i < config.display.video_size / 1024; i++) {
		static char dummy[1024];

		ret = write(fd, dummy, sizeof(dummy));
		assert(ret == sizeof(dummy));
	}

	ret = dup2(fd, FD_VIDEO_MEMORY);
	assert(ret == FD_VIDEO_MEMORY);

	ret = close(fd);
	assert(0 <= ret);

	if (config.display.video_mode /* &&
	   !sync_with_frontend(config.display.video_mode) */) {
		vsd = (video_struct *) malloc(sizeof(video_struct));
		if (!vsd) {
			STDERR_MSG("video simulation data allocation failure");
			exit(1);
		}

		vsd->vid_map = (char *) mmap(0, config.display.video_size,
				PROT_READ | PROT_WRITE, MAP_SHARED,
				FD_VIDEO_MEMORY, 0x00000000);
		assert(0 < (int)vsd->vid_map);

		vsd->pal_map = (char *) mmap(0, SZ_VIDEO_CHIP,
				PROT_READ | PROT_WRITE, MAP_SHARED,
				FD_VIDEO_CHIP, 0x0000000);
		assert(0 < (int) vsd->pal_map);

		if (setup_video_sim(config.display.video_mode)) {
			STDERR_MSG("quit video sim init");
			exit(1);
		}

		switch(vsd->pid = fork()) {
		case -1:
			STDERR_MSG("umvideo oops");
			exit(-1);
		case 0:
			memset(vsd->layers[0], 0, vsd->vmap_size);
			memset(vsd->layers[1], 0, vsd->vmap_size);

			while(1) {
				sat = vsd->update_palette();
				vsd->do_video(sat);
			}
		}
	}
}

void
umvideo_shutdown(void)
{
	int ret;
	int status;

	if (vsd) {
		ret = kill(vsd->pid, SIGKILL);
		assert(0 <= ret);

		ret = waitpid(vsd->pid, &status, 0);
		assert(ret == vsd->pid);
		assert(WIFSIGNALED(status) && WTERMSIG(status) == SIGKILL);
	}

	// UMLfront is no longer used. So the statement below is no longer
	// required. I am commenting it as it causes a segfault on guest
	// shutdown. Ideally the whole file should be reviewed and all
	// unnecessary code should be removed -MAB

	//vt_packet[0] = 8;
	//send_video_packet(vt_packet);

	ret = close(FD_VIDEO_MEMORY);
	assert(0 <= ret);

	ret = close(FD_VIDEO_CHIP);
	assert(0 <= ret);
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
