
#include <string.h>
#include "decode.h"

struct efl {
	unsigned c:1, b1:1, p:1, b3:1, b4:1, b5:1, z:1, s:1, b8:1,
		 b9:1, b10:1, o:1;  
};

#define CF(x) static int x(struct efl efl,unsigned ecx)

CF(c_o) { return efl.o; }
CF(c_c) { return efl.c; }
CF(c_z) { return efl.z; }
CF(c_a) { return efl.c==0 && efl.z==0; }
CF(c_s) { return efl.s; }
CF(c_p) { return efl.p; }
CF(c_l) { return efl.s!=efl.o; }
CF(c_g) { return efl.z==0 && efl.s==efl.o; }
CF(c_cxz) { return ecx==0; }
CF(c_loop) { return ecx!=1; }
CF(c_loopz) { return ecx!=1 && efl.z==1; }
CF(c_loopnz) { return ecx!=1 && efl.z==0; }

struct opdef { 
	unsigned char type;
	unsigned char neg;
	int (*cond)(struct efl fl,unsigned ecx);
};

#define SIMPLE { JUMP_UNCON }
#define COND(x,y) { JUMP_COND, y, c_##x }

#define JUMP_COND  1
#define JUMP_UNCON 2

static struct opdef onebyte_jumps[256]= {
/* CALL */
[0xe8] SIMPLE,
[0x9a] SIMPLE,

/* Jcc rel8 */
[0x70] COND(o, 0),
[0x71] COND(o, 1),
[0x72] COND(c, 0),
[0x73] COND(c, 1),
[0x74] COND(z, 0),
[0x75] COND(z, 1),
[0x76] COND(a, 1),
[0x77] COND(a, 0),
[0x78] COND(s, 0),
[0x79] COND(s, 1),
[0x7a] COND(p, 0),
[0x7b] COND(p, 1),
[0x7c] COND(l, 0),
[0x7d] COND(l, 1),
[0x7e] COND(g, 1),
[0x7f] COND(g, 0),
[0xe3] COND(cxz,0),

/* JMP */
[0xeb] SIMPLE,
[0xe9] SIMPLE,
[0xea] SIMPLE,

/* RET */
[0xc3] SIMPLE,
[0xcb] SIMPLE,
[0xc2] SIMPLE,
[0xca] SIMPLE,

/* LOOP */
[0xe2] COND(loop, 0),
[0xe1] COND(loopz, 0),
[0xe0] COND(loopnz, 0),

/* INT */
[0xcd] SIMPLE
}; 

static struct opdef twobyte_jumps[256]= {
	
/* Jcc rel16/32 */
[0x80] COND(o, 0),
[0x81] COND(o, 1),
[0x82] COND(c, 0),
[0x83] COND(c, 1),
[0x84] COND(z, 0),
[0x85] COND(z, 1),
[0x86] COND(a, 1),
[0x87] COND(a, 0),
[0x88] COND(s, 0),
[0x89] COND(s, 1),
[0x8a] COND(p, 0),
[0x8b] COND(p, 1),
[0x8c] COND(l, 0),
[0x8d] COND(l, 1),
[0x8e] COND(g, 1),
[0x8f] COND(g, 0),

};

static struct opdef ffext_jumps[8]= {
	/* CALL */
	[2] SIMPLE,
	[3] SIMPLE,
	/* JMP */
	[4] SIMPLE,
	[5] SIMPLE,
};

#define MODRM(x)  ((x>>3) & 7)
#define mode32 0
#define mode16 1


#define PFX_LOCK   (1<<0)
#define PFX_REPNE  (1<<1)
#define PFX_REPE   (1<<2)
#define PFX_CS     (1<<3)
#define PFX_SS     (1<<4)
#define PFX_DS     (1<<5)
#define PFX_ES     (1<<6)
#define PFX_FS     (1<<7)
#define PFX_GS     (1<<8)
#define PFX_OSIZE  (1<<9)
#define PFX_ASIZE  (1<<10)


int parse_prefixes(unsigned char *code,int *cstart) {
	int i;
	int pfx;
	unsigned char *f;
	unsigned char prefix_list[]= { 
	  0xf0, 0xf2, 0xf3, 
	  0x2e, 0x36, 0x3e, 0x26, 0x64, 0x65, 0x66, 0x67 };
	pfx= 0;
	for(i=0;i<15;i++) {
		int j;
		f= memchr(prefix_list,code[i],sizeof(prefix_list));
		if (!f)
			break;
		j= f-prefix_list;
		pfx= pfx | (1<<j);
	}
	*cstart= i;
	return pfx;
}

static int is_string(unsigned char *code) {
	static const unsigned char strinsts[]= {
	  0xa4, 0xa5, 0xa6, 0xa7,
	  0x6c, 0x6d, 0x6e, 0x6f,
	  0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf };
	return memchr(strinsts,code[0],sizeof(strinsts))!=0;
}



int is_syscall(unsigned char *code) {
	return code[0]==0xcd && code[1]==0x80;
}

int is_repstring(unsigned char *code) {
	int cstart;
	int pfx;
	pfx= parse_prefixes(code, &cstart);
	if (!(pfx & (PFX_REPNE | PFX_REPE))) 
		return 0;
	code += cstart;
	return is_string(code);
}


int branch_taken(unsigned char *code,int _efl,unsigned ecx) {
	unsigned char c;
	struct opdef *z;	
	int y;
	int mode=mode32;
	struct efl efl;
	volatile union {
		struct efl efl;
		int _efl;
	} flg;

	flg._efl= _efl;
	efl= flg.efl;

again:
	c= *(code++);
	switch(c) {
	  case 0x66: mode= mode16; goto again;
	  case 0xff: z= ffext_jumps+ MODRM(*code) ; break;
	  case 0x0f: z= twobyte_jumps + *code; break;
	  default: z= onebyte_jumps+c;
	}

	if (mode==mode16) ecx &=  0xffff;

	switch (z->type) {
	  case JUMP_UNCON: 
		return 1;
	  case JUMP_COND: 
		y= (*z->cond)(efl,ecx);
		return (y && !z->neg) || (!y && z->neg); /* logical XOR */
	  default:
		return 0;
	}
}

