#include <stdio.h>
#include <revirt/rlog.h>
#include <revirt/flags.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include "printlog.h"

static const char *type_string[]= {
#define X(x) [ENTRY_##x] #x
  X(SCHED),
  X(FORK),
  X(EXIT),
  X(TOP),
  X(SYSRET),
#undef X
};

void print_schedlog(struct rlog_schedule_entry *e) {
	printf("type:%s pid:%d flags:%x\nx:%016llx ix:%08x pf:%08x dev:%08x d:%08x\n",
	       type_string[e->type], e->pid, e->flags, e->ctr.hw_branch, 
	       e->ctr.hw_intr, e->ctr.ex_pf, e->ctr.ex_dev);
	printf("eip:0x%x\n", e->eip);
}

int ztrace(int command,int pid,unsigned long addr,void *data) {
      int ret;
      ret= syscall(__NR_ztrace, command, pid, addr, data);
      return ret;
}

void print_replayflags(unsigned int replayflags) {
	printf("replay_flags: 0x%x (", replayflags);
	if(replayflags & REPLAY_LOG)
		printf("LOG ");
	if(replayflags & REPLAY_REPLAY)
		printf("REPLAY ");
	if(replayflags & REPLAY_SSTEP)
		printf("SSTEP ");
	if(replayflags & REPLAY_START)
		printf("START ");
	if(replayflags & REPLAY_PROCESS_IN_UMODE)
		printf("PROCESS_IN_UMODE ");
	if(replayflags & REPLAY_SCHEDPEND)
		printf("SCHEDPEND ");
	if(replayflags & REPLAY_SYNC)
		printf("SYNC ");
	if(replayflags & REPLAY_WAIT)
		printf("WAIT ");
	if(replayflags & REPLAY_AT_EXIT)
		printf("AT_EXIT ");
	if(replayflags & REPLAY_NOTDONE)
		printf("NOTDONE ");
	if(replayflags & REPLAY_SCHED)
		printf("SCHED ");
	if(replayflags & REPLAY_IN_COPY_USER)
		printf("IN_COPY_USER ");
	if(replayflags & REPLAY_DATASQUASH)
		printf("DATASQUASH ");
	if(replayflags & REPLAY_XUSER)
		printf("XUSER ");
	if(replayflags & REPLAY_ZOMBIE)
		printf("ZOMBIE ");
	if(replayflags & REPLAY_SYSCALL)
		printf("SYSCALL ");
	if(replayflags & REPLAY_PTRACE)
		printf("PTRACE ");
	if(replayflags == 0)
		printf("NONE ");
	printf(")\n");
}

void print_ztflags(unsigned int zt_flags) {
	printf("zt_flags: 0x%x (", zt_flags);
	if(zt_flags & ZTRACE_XUSER_STOP)
		printf("XUSER_STOP ");
	if(zt_flags & ZTRACE_SYSENTER_STOP)
		printf("SYSENTER_STOP ");
	if(zt_flags & ZTRACE_SYSEXIT_STOP)
		printf("SYSEXIT_STOP ");
	if(zt_flags & ZTRACE_BKPT)
		printf("BKPT ");
	if(zt_flags & ZTRACE_BKPT_IN_KERNEL)
		printf("BKPT_IN_KERNEL ");
	if(zt_flags & ZTRACE_WAITEXIT)
		printf("WAITEXIT ");
	if(zt_flags & ZTRACE_BLOCK_STOP)
		printf("BLOCK_STOP ");
	if(zt_flags & ZTRACE_SIGNAL_STEP)
		printf("SIGNAL_STEP ");
	if(zt_flags & ZTRACE_PERF)
		printf("PERF ");
	printf(")\n");
}

void print_state(struct ztrace_state_struct * zt_s) {
    	int i;
#define zt_state (*zt_s)
	printf( "eax:0x%8x\teip:0x%8x\n" \
		"ebx:0x%8x\tesp:0x%8x\n" \
		"ecx:0x%8x\tebp:0x%8x\n" \
		"edx:0x%8x\txds:0x%8x\n" \
		"esi:0x%8x\txes:0x%8x\n" \
		"edi:0x%8x\txcs:0x%8x\n" \
		"efl:0x%8x\txss:0x%8x\n" \
		"orig_eax: 0x%8x\n",
		zt_state.regs.eax, zt_state.regs.eip,
		zt_state.regs.ebx, zt_state.regs.esp,
		zt_state.regs.ecx, zt_state.regs.ebp,
		zt_state.regs.edx, zt_state.regs.xds,
		zt_state.regs.esi, zt_state.regs.xes,
		zt_state.regs.edi, zt_state.regs.xcs,
		zt_state.regs.eflags, zt_state.regs.xss,
		zt_state.regs.orig_eax );
	print_replayflags(zt_state.replay_flags);
	print_ztflags(zt_state.zt_flags);
	printf( "zt_rununtil: eip:0x%x\twchan:0x%x\tflags:0x%08x\n",
		zt_state.zt_rununtil.eip,
		zt_state.zt_rununtil.wchan,
		zt_state.zt_rununtil.flags);
	printf("cr2:0x%8x\ttrap_no:0x%8x\terror_code:0x%8x\n",
		zt_state.cr2, zt_state.trap_no, zt_state.error_code);
	for(i=0; i<8; i++) {
		printf("debugreg[%d]:0x%x\n", i, zt_state.debugreg[i]);
	}
	printf("code: ");
	for(i=0; i<16; i++) {
		printf("%x ", (int) zt_state.code[i]);
	}
	printf("\npcode: ");
	for(i=0; i<16; i++) {
		printf("%x ", (int) zt_state.pcode[i]);
	}
	printf("\n");
	printf("count: 0x%llx [%llx %x %x %x]\n",
		ADJ_BC(zt_state.ctr),
		zt_state.ctr.hw_branch,
		zt_state.ctr.hw_intr,
		zt_state.ctr.ex_pf,
		zt_state.ctr.ex_dev);
	printf("zt_perferr: %llx\n", zt_state.zt_perferr);
#undef zt_state
}
